/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

 This file is part of the TerraLib - a Framework for building GIS enabled applications.

 TerraLib is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License,
 or (at your option) any later version.

 TerraLib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with TerraLib. See COPYING. If not, write to
 TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
 \file terralib/itest/raster/ItsSpatialIndex.cpp

 \brief A integration test suit for the TerraLib VP Module.

 \author Mario Rocco Pettinati
 */

// TerraLib
#include <terralib_buildconfig.h>

// TerraLib
#include <terralib/core/uri/URI.h>
#include <terralib/dataaccess/utils/DataAccess.h>
#include <terralib/vp/AbstractOperation.h>
#include <terralib/vp/OperationThreadManager.h>
#include <terralib/vp/SubdivideDataHandler.h>

#include "TestUtils.h"

// GTest and GMock
#include <gmock/gmock.h>
#include <gtest/gtest.h>

namespace
{
  class ITsOperations : public ::testing::Test
  {
    virtual void SetUp()
    {
    }

    virtual void TearDown()
    {
    }
  };

  void execute(te::vp::AbstractOperation* operation)
  {
    te::gtest::DataInfo redDataInfo = te::gtest::GetRedDataInfo();
    te::gtest::DataInfo blueDataInfo = te::gtest::GetBlueDataInfo();

    //loads the input data access
    std::unique_ptr<te::da::DataAccess> inputRedDataAccess(new te::da::DataAccess(te::core::URI(redDataInfo.m_inputUri), "OGR", redDataInfo.m_inputDatasetName));
    std::unique_ptr<te::da::DataAccess> inputBlueDataAccess(new te::da::DataAccess(te::core::URI(blueDataInfo.m_inputUri), "OGR", blueDataInfo.m_inputDatasetName));

    //creates the output
    std::unique_ptr<te::da::DataSetType> outputDataSetType(static_cast<te::da::DataSetType*>(inputRedDataAccess->getDataSetType()->clone()));
    std::string outputDataSetName = operation->getName() + "_" + redDataInfo.m_inputDatasetName + "_" + blueDataInfo.m_inputDatasetName;
    std::string outputFileName = redDataInfo.m_outputDir + "/" + outputDataSetName + ".shp";
    if (te::core::FileSystem::exists(outputFileName) == true)
    {
      te::core::FileSystem::remove(outputFileName);
    }

    std::unique_ptr<te::da::DataAccess> outputDataAccess(new te::da::DataAccess(te::core::URI(outputFileName), "OGR", outputDataSetName));

    te::vp::AbstractParameters* params = new te::vp::AbstractParameters();
    params->addInputDataAccess(inputRedDataAccess.get());
    params->addInputDataAccess(inputBlueDataAccess.get());
    params->addOutputDataAccess(outputDataAccess.get());
    params->setMaxCoordinates(1000);

    operation->setParameters(params);
    te::vp::AbstractDataHandler* dataHandler = new te::vp::SubdivideDataHandler();

    te::vp::OperationThreadManager threadManager(operation, dataHandler);
    threadManager.start();
  }

  TEST_F(ITsOperations, Difference)
  {
    te::vp::AbstractOperation* operation = te::vp::OperationFactory::make("Difference");
    execute(operation);
  }

  TEST_F(ITsOperations, Identity)
  {
    te::vp::AbstractOperation* operation = te::vp::OperationFactory::make("Identity");
    execute(operation);
  }

  TEST_F(ITsOperations, Intersection)
  {
    te::vp::AbstractOperation* operation = te::vp::OperationFactory::make("Intersection");
    execute(operation);
  }

  TEST_F(ITsOperations, Union)
  {
    te::vp::AbstractOperation* operation = te::vp::OperationFactory::make("Union");
    execute(operation);
  }
}
 
