/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/geometry/Validation.cpp

  \brief Utility functions for the Geometry Module.
*/

// TerraLib
#include "Validation.h"

#include "../common/STLUtils.h"
#include "../core/logger/Logger.h"
#include "CommonDataStructures.h"
#include "CoordinateSnapper.h"
#include "Curve.h"
#include "GEOSReader.h"
#include "GEOSWriter.h"
#include "GeometryCollection.h"
#include "GeometryFunctions.h"
#include "LinearRing.h"
#include "LineString.h"
#include "MultiLineString.h"
#include "Polygon.h"
#include "Utils.h"
#include "validation/AddProjectedCoordinates.h"
#include "validation/FixNoding.h"
#include "validation/FixOrientation.h"
#include "validation/FixRepeatedCoordinates.h"
#include "validation/MergeCoordinates.h"
#include "validation/RemoveCollapsedSegments.h"

//Geos
#include <geos/geom/Geometry.h>
#include <geos/geom/Polygon.h>
#include <geos/operation/polygonize/Polygonizer.h>
#include <geos/operation/valid/IsValidOp.h>
#include <geos/util/GEOSException.h>

//STL
#include <vector>

bool te::gm::Validation::CheckValidity(const te::gm::Geometry* geom, te::gm::TopologyValidationError& error)
{
#ifdef TERRALIB_GEOS_ENABLED

  try
  {
    std::unique_ptr<geos::geom::Geometry> g(GEOSWriter::write(geom));

    geos::operation::valid::IsValidOp vop(g.get());

    geos::operation::valid::TopologyValidationError* err = vop.getValidationError();

    if (err)
    {
      error.m_coordinate = Coord2D(err->getCoordinate().x, err->getCoordinate().y);
      error.m_message = err->getMessage();

      return false;
    }

    return true;
  }
  catch(const geos::util::GEOSException& e)
  {
    error.m_message = e.what();

    return false;
  }

#else
  throw Exception(TE_TR("isValid routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::Geometry* te::gm::Validation::Validate(te::gm::Geometry* geom)
{
  if (te::gm::IsNullOrEmpty(geom) == true)
  {
    return nullptr;
  }

#ifdef TERRALIB_GEOS_ENABLED

  std::unique_ptr<geos::geom::Geometry> g(GEOSWriter::write(geom));

  te::gm::GeometryVector pAdd;
  geos::operation::polygonize::Polygonizer polygonizer;

  switch (g->getGeometryTypeId())
  {
    case geos::geom::GEOS_POLYGON:
    {
      if (g->isValid())
      {
        g->normalize(); // validate does not pick up rings in the wrong order - this will fix that
        return GEOSReader::read(g.get()); // If the polygon is valid just return it
      }

      AddPolygon(dynamic_cast<te::gm::Polygon*>(GEOSReader::read(g.get())), pAdd);
    }
    break;

    case geos::geom::GEOS_MULTIPOLYGON:
    {
      if (g->isValid()){
        g->normalize(); // validate does not pick up rings in the wrong order - this will fix that
        return GEOSReader::read(g.get()); // If the multipolygon is valid just return it
      }

      for (std::size_t n = g->getNumGeometries(); n-- > 0;)
        AddPolygon(dynamic_cast<te::gm::Polygon*>(GEOSReader::read(g->getGeometryN(n))), pAdd);

    }
    break;

    default:
      return static_cast<te::gm::Geometry*>(geom->clone());
  }

  for (std::size_t i = 0; i < pAdd.size(); i++)
    polygonizer.add(GEOSWriter::write(pAdd[i]));

  te::common::FreeContents(pAdd);

  std::vector <geos::geom::Polygon*>* vecpolGeos = polygonizer.getPolygons();

  switch (vecpolGeos->size())
  {
    case 1:
    {
      vecpolGeos->at(0)->setSRID(geom->getSRID());

      te::gm::Geometry* teGeom = GEOSReader::read(vecpolGeos->at(0)); // single polygon - no need to wrap

      te::common::FreeContents(*vecpolGeos);
      delete vecpolGeos;

      return teGeom; 
    }
    default:
    {
      //filter the invalid polygons
      std::size_t i = 0;
      while (i < vecpolGeos->size())
      {
        if (!vecpolGeos->at(i)->isValid())
        {
          delete vecpolGeos->at(i);
          vecpolGeos->erase(vecpolGeos->begin() + static_cast<int>(i));
        }
        else
        {
          ++i;
        }
      }
      te::gm::Geometry* ret = nullptr;

      if (vecpolGeos->empty() == false)
      {
        //polygons may still overlap! Need to sym difference them
        vecpolGeos->at(0)->setSRID(geom->getSRID());
        ret = GEOSReader::read(vecpolGeos->at(0));

        for (std::size_t p = 1; p < vecpolGeos->size(); p++)
        {
          vecpolGeos->at(p)->setSRID(geom->getSRID());
          ret = ret->symDifference(GEOSReader::read(vecpolGeos->at(p)));
        }

        te::common::FreeContents(*vecpolGeos);
      }

      delete vecpolGeos;

      return ret;
    }
  }

#else
  throw Exception(TE_TR("validate routine is supported by GEOS! Please, enable the GEOS support."));
#endif
}

te::gm::GeometryPtr te::gm::Validation::MakeValid(const te::gm::Geometry* geometry)
{
  std::string errorMessage;
  return MakeValid(geometry, errorMessage);
}

te::gm::GeometryPtr te::gm::Validation::MakeValid(const te::gm::Geometry* geometry, std::string& errorMessage)
{
  te::gm::GeometryPtr result(nullptr);
  if (te::gm::IsNullOrEmpty(geometry) == true)
  {
    return result;
  }

  double tolerance = te::gm::GetAppropriatePrecision(geometry->getSRID());
  te::gm::GeomType simpleGeomType = te::gm::GetSimpleType(geometry->get2DGeomTypeId());

  try
  {
    //1 - Removes all the repeated consecutive coordidates
    AbstractValidatorPtr fixRepeatedCoordinates(new FixRepeatedCoordinates());
    result.reset(fixRepeatedCoordinates->validate(geometry).release());

    //2 - Fix the orientation of the rings of the polygon, when appliable
    AbstractValidatorPtr fixOrientation(new FixOrientation());
    result.reset(fixOrientation->validate(result.get()).release());

    //3 - Add perpendicular points that are within the tolerance to each analysed segment
    AbstractValidatorPtr addProjectedCoordinates(new AddProjectedCoordinates());
    addProjectedCoordinates->setPrecision(tolerance);
    result.reset(addProjectedCoordinates->validate(result.get()).release());

    //4 - We snap the coordinates to ensure that any "similar" coordinate is now equal
    AbstractValidatorPtr mergeCoordinates(new MergeCoordinates());
    mergeCoordinates->setPrecision(tolerance);
    result.reset(mergeCoordinates->validate(result.get()).release());

    //4.1 - Fix the orientation of the rings of the polygon, when appliable
    result.reset(fixRepeatedCoordinates->validate(result.get()).release());

    //5 - To avoid the discard of rings in geos polygonize function, we must ensure that we removed all the collapsed segments
    AbstractValidatorPtr removeCollapsedSegments(new RemoveCollapsedSegments());
    result.reset(removeCollapsedSegments->validate(result.get()).release());
    
    //6 - We finally fix the noding of the geometry
    AbstractValidatorPtr fixNoding(new FixNoding());
    result.reset(fixNoding->validate(result.get()).release());
    
    //7 - and finally we filter the result so we can ensure to return the same geometric type
    result.reset(te::gm::NormalizeGeometryByType(result.get(), simpleGeomType));
  }
  catch (const std::exception& e)
  {
    errorMessage = "Error fixing geometry: " + std::string(e.what());
    TE_LOG_WARN(errorMessage);

    //re-throw the exception
    return result;
  }

  if (simpleGeomType != te::gm::PolygonType)
  {
    return result;
  }

  //now we validate if the area of the polygon has changed significantly
  double originalArea = te::gm::GetArea(geometry);
  double fixedArea = te::gm::GetArea(result.get());

  if (std::fabs(originalArea - fixedArea) >= tolerance)
  {
    std::string originalWKT = geometry->toString();

    std::string logMessage = "Function: MakeValid";
    logMessage += "\nDescription: The area of the output geometry has changed significantly in comparisson to the area of the input geometry";
    logMessage += "\nWKT Input: " + originalWKT;

    if (te::gm::IsNullOrEmpty(result.get()) == false)
    {
      std::string fixedWKT = result->toString();
      logMessage += "\nWKT Output: " + fixedWKT;
    }
    else
    {
      logMessage += "\nWKT Output: Empty";
    }

    errorMessage = logMessage;

    TE_LOG_WARN(logMessage);
  }

  return result;
}
