/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/geometry/Validation.h

  \brief A list of Validation functions for the Geometry Module.
*/

#ifndef __TERRALIB_GEOMETRY_INTERNAL_VALIDATION_H
#define __TERRALIB_GEOMETRY_INTERNAL_VALIDATION_H

// TerraLib
#include "Config.h"

#include "Coord2D.h"

// STL
#include <map>
#include <string>
#include <vector>

namespace te
{
  namespace gm
  {
// Forward declarations
    class Geometry;
    class GeometryPtr;
    class LineString;
   
    /*!
    \struct TopologyValidationError

    \brief This struct contains informations about GEOS TopologyValidationError
    */
    struct TopologyValidationError
    {
      public:

        Coord2D     m_coordinate;
        std::string m_message;
    };

    class TEGEOMEXPORT Validation
    {
    public:

      /*!
        \brief It check geometry validity using GEOS.

        \param geom  Geometry that will be verified.
        \param error TopologyValidationError struct.

        \return True if geometry is valid.
      */
      static bool CheckValidity(const Geometry* geom, te::gm::TopologyValidationError& error);

      /*!
            \brief Get/create a valid version of the geometry given. If the geometry is a polygon or multi polygon, self intersections /
                   inconsistencies are fixed. Otherwise the geometry is returned.

            \param geom
            \return a geometry

            \note: https://stackoverflow.com/questions/31473553/is-there-a-way-to-convert-a-self-intersecting-polygon-to-a-multipolygon-in-jts
            */
      static te::gm::Geometry* Validate(te::gm::Geometry* geom);

      /*!
        \brief Apply several fixes to a geometry to ensure that it will be made valid.

        \description It applies the following fixes:
                     1 - Removes all the repeated consecutive coordidates
                     2 - Add perpendicular points that are within the tolerance to each analysed segment
                     3 - We snap the coordinates to ensure that any "similar" coordinate is now equal
                     4 - To avoid the discard of rings in geos polygonize function, we must ensure that we removed all the collapsed segments
                     5 - Finally we applies a polygonizer to recreate the geometry and ensure that it is now consistent

        \param geom  Geometry that will be verified.

        \return True if geometry is valid.
      */
      static te::gm::GeometryPtr MakeValid(const te::gm::Geometry* geometry);

      static te::gm::GeometryPtr MakeValid(const te::gm::Geometry* geometry, std::string& errorMessage);
    };

  } // end namespace gm
}   // end namespace te

#endif  // __TERRALIB_GEOMETRY_INTERNAL_VALIDATION_H
