#include "FixNoding.h"

#include "../../common/STLUtils.h"
#include "../Geometry.h"
#include "../GeometryCollection.h"
#include "../LineString.h"
#include "../Polygon.h"
#include "../Utils.h"
#include "../GEOSGeometryFactory.h"
#include "../GEOSReader.h"
#include "../GEOSWriter.h"

#include <geos/algorithm/LineIntersector.h>
#include <geos/geom/CoordinateSequence.h>
#include <geos/geom/Geometry.h>
#include <geos/geom/LinearRing.h>
#include <geos/geom/LineString.h>
#include <geos/geom/Polygon.h>
#include <geos/noding/IntersectionAdder.h>
#include <geos/noding/MCIndexNoder.h>
#include <geos/noding/NodedSegmentString.h>
#include <geos/operation/polygonize/Polygonizer.h>

te::gm::GeometryVector GetFragments(const te::gm::Polygon* polygon, bool& wasFragmented)
{
  wasFragmented = false;
  std::unique_ptr<geos::geom::Geometry> geosGeom(te::gm::GEOSWriter::write(static_cast<const te::gm::Geometry*>(polygon)));
  geos::geom::Polygon* geosPolygon = dynamic_cast<geos::geom::Polygon*>(geosGeom.get());

  std::vector<geos::noding::SegmentString*> vecSegmentString;

  //we first add the exterior ring
  geos::geom::CoordinateSequence* coordinateSequence = geosPolygon->getExteriorRing()->getCoordinates();
  geos::noding::SegmentString* segmentString = new geos::noding::NodedSegmentString(coordinateSequence, nullptr);
  vecSegmentString.push_back(segmentString);

  //then we add the interior rings
  for (std::size_t i = 0; i < geosPolygon->getNumInteriorRing(); ++i)
  {
    geos::geom::CoordinateSequence* coordinateSequence = geosPolygon->getInteriorRingN(i)->getCoordinates();
    geos::noding::SegmentString* segmentString = new geos::noding::NodedSegmentString(coordinateSequence, nullptr);
    vecSegmentString.push_back(segmentString);
  }

  geos::algorithm::LineIntersector lineIntersector(geosPolygon->getPrecisionModel());
  geos::noding::MCIndexNoder noder(new geos::noding::IntersectionAdder(lineIntersector));

  noder.computeNodes(&vecSegmentString);
  std::vector<geos::noding::SegmentString*>* vecNodedSegmentString = noder.getNodedSubstrings();

  te::gm::GeometryVector vecFragments;
  for (std::size_t i = 0; i < vecNodedSegmentString->size(); ++i)
  {
    geos::geom::CoordinateSequence* geosCoordSequence = vecNodedSegmentString->at(i)->getCoordinates();
    geos::geom::Geometry* geosFragment = te::gm::GEOSGeometryFactory::getGeomFactory()->createLineString(geosCoordSequence);
    te::gm::Geometry* fragment = te::gm::GEOSReader::read(geosFragment);
    fragment->setSRID(polygon->getSRID());

    vecFragments.push_back(fragment);
  }

  if (vecSegmentString.size() != vecFragments.size())
  {
    wasFragmented = true;
  }

  te::common::FreeContents(vecSegmentString);
  te::common::FreeContents(*vecNodedSegmentString);
  delete vecNodedSegmentString;

  return vecFragments;
}

te::gm::GeometryPtr te::gm::FixNoding::validatePolygon(const te::gm::Polygon* polygon, bool& wasChanged) const
{
  wasChanged = false;

  te::gm::GeometryPtr result;
  try
  {
    result.reset(polygon->buffer(0));
  }
  catch (const std::exception&)
  {
    te::gm::GeometryPtr result;
    return result;
  }

  double originalArea = te::gm::GetArea(polygon);
  double fixedArea = te::gm::GetArea(result.get());

  double minAreaChange = te::gm::GetAppropriatePrecision(polygon->getSRID());

  //after fixing the noding, we must check if we had significant area change.
  if (std::fabs(originalArea - fixedArea) >= minAreaChange)
  {
    try
    {
      //if so, we try to reconstruct the polygon based of the segments
    //the first step is to fragment all the segments using the self intersection points
      bool wasFragmented = false;
      te::gm::GeometryVector vecFragments = GetFragments(polygon, wasFragmented);
      if (wasFragmented == false)
      {
        result.reset(static_cast<te::gm::Geometry*>(polygon->clone()));
        return result;
      }

      wasChanged = true;

      //then we get all the segments and polygonize them
      te::gm::GeometryVector vecPolygonized;
      te::gm::Polygonizer(vecFragments, vecPolygonized);

      te::gm::GeometryVector vecPolygonizedCleaned;
      for (std::size_t i = 0; i < vecPolygonized.size(); ++i)
      {
        te::gm::Polygon* fragmentPolygon = static_cast<te::gm::Polygon*>(vecPolygonized.at(i));
        te::gm::Coord2D interiorCoord = te::gm::GetInteriorPoint(fragmentPolygon);
        if (te::gm::IsPointOnPolygon(interiorCoord, polygon) == true)
        {
          vecPolygonizedCleaned.push_back(static_cast<te::gm::Geometry*>(fragmentPolygon->clone()));
        }
      }

      te::common::FreeContents(vecPolygonized);

      result.reset(te::gm::CreateCollectionFromGeometryVector(vecPolygonizedCleaned).release());

      te::common::FreeContents(vecPolygonizedCleaned);
      te::common::FreeContents(vecFragments);

      double newFixedArea = te::gm::GetArea(result.get());

      //after fixing the noding, we must check if we had significant area change.
      if (std::fabs(originalArea - newFixedArea) >= minAreaChange)
      {
        std::string wktError = result->toString();
      }
    }
    catch (const std::exception&)
    {
      return result;
    }
  }

  return result;
}

te::gm::GeometryPtr te::gm::FixNoding::validateLine(const te::gm::LineString* lineString, bool& wasChanged) const
{
  wasChanged = false;
  te::gm::GeometryPtr result(static_cast<te::gm::Geometry*>(lineString->clone()));
  return result;
}
