/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/qt/widgets/rp/RasterEditWizardPage.cpp

  \brief This file defines a class for a RasterEdit Wizard page.
*/

// TerraLib
#include "../../../common/StringUtils.h"
#include "../../../dataaccess/dataset/DataSet.h"
#include "../../../dataaccess/dataset/ObjectIdSet.h"
#include "../../../dataaccess/utils/Utils.h"
#include "../../../geometry/Geometry.h"
#include "../../../geometry/GeometryProperty.h"
#include "../../../geometry/GeometryCollection.h"
#include "../../../geometry/LinearRing.h"
#include "../../../geometry/Polygon.h"
#include "../../../geometry/Utils.h"
#include "../../../raster/Grid.h"
#include "../../../raster/Raster.h"
#include "../../../maptools/Utils.h"
#include "../canvas/Canvas.h"
#include "../canvas/MapDisplay.h"
#include "RpToolsWidget.h"

#include "RasterEditWizardPage.h"
#include "ui/ui_RasterEditWizardPageForm.h"

// Qt
#include <QGridLayout>
#include <QCheckBox>
#include <QDoubleValidator>
#include <QIntValidator>
#include <QMessageBox>
#include <QTableWidgetItem>

// STL
#include <memory>

Q_DECLARE_METATYPE(te::map::AbstractLayerPtr)

te::qt::widgets::RasterEditWizardPage::RasterEditWizardPage(QWidget* parent)
  : QWizardPage(parent),
    m_ui(new Ui::RasterEditWizardPageForm),
    m_layer(nullptr),
    m_mapDisplay(nullptr)
//    m_count(0)
{
// setup controls
  m_ui->setupUi(this);

  m_ui->m_editionTypeComboBox->clear();
  m_ui->m_editionTypeComboBox->addItem(tr("Entire raster"), EDITION_ENTIRE_RASTER);
  m_ui->m_editionTypeComboBox->addItem(tr("ROI assisted"), EDITION_ROI_ASSISTED);

//build form
  QGridLayout* displayLayout = new QGridLayout(m_ui->m_navigatorWidget);
  m_navigator.reset( new te::qt::widgets::RpToolsWidget(m_ui->m_navigatorWidget));
  displayLayout->addWidget(m_navigator.get());
  displayLayout->setContentsMargins(0,0,0,0);

  m_ui->m_removeRoiToolButton->setIcon(QIcon::fromTheme("edit-deletetool"));

//configure page
  this->setTitle(tr("Raster Edit"));
  this->setSubTitle(tr("Select the type of EDITION and set their specific parameters."));

//connects

  connect(m_ui->m_roiTableWidget, SIGNAL(cellClicked(int, int)), this, SLOT(onSelectRoiItem()));
  connect(m_ui->m_removeRoiToolButton, SIGNAL(clicked()), this, SLOT(onRemoveRoiItem()));

  connect(m_ui->m_addValuesPushButton, SIGNAL(clicked()), this, SLOT(addValuesPushButtonClicked()));
  connect(m_ui->m_removeValuesPushButton, SIGNAL(clicked()), this, SLOT(removeValuesPushButtonClicked()));

  connect(m_navigator.get(), SIGNAL(envelopeAcquired(te::gm::Envelope)), this, SLOT(onEnvelopeAcquired(te::gm::Envelope)));
  connect(m_navigator.get(), SIGNAL(geomAquired(te::gm::Polygon*)), this, SLOT(onGeomAquired(te::gm::Polygon*)));
}

te::qt::widgets::RasterEditWizardPage::~RasterEditWizardPage()
{
  const std::size_t geomVecSize = m_geomVec.size();
  for( std::size_t geomVecIdx = 0 ; geomVecIdx < geomVecSize ; ++geomVecIdx )
  {
    delete m_geomVec[ geomVecIdx ];
  }
}

bool te::qt::widgets::RasterEditWizardPage::isComplete() const
{
  int index = m_ui->m_editionTypeComboBox->currentIndex();
  int type = m_ui->m_editionTypeComboBox->itemData(index).toInt();

  if (type == EDITION_ENTIRE_RASTER)
  {
    if( m_ui->m_roiSubstitutionValuesTableWidget->rowCount() == 0 )
      return false;
    else
      return true;
  }
  else if(type == EDITION_ROI_ASSISTED)
  {
    if(
        ( m_ui->m_roiSubstitutionValuesTableWidget->rowCount() == 0 )
        ||
        ( m_ui->m_roiTableWidget->rowCount() == 0 )
      )
      return false;
    else
      return true;
  }
  else
    return false;
}

void te::qt::widgets::RasterEditWizardPage::setLayer(te::map::AbstractLayerPtr layer)
{
  m_layer = layer;

  // bands combo

  m_ui->m_bandComboBox->clear();

  if( layer.get() )
  {
    std::unique_ptr<te::rst::Raster> inputRst(te::map::GetRaster(layer.get()));

    for( unsigned int band = 0 ; band < inputRst->getNumberOfBands() ; ++band )
    {
      m_ui->m_bandComboBox->addItem( QString::number( band ) );
    }
  }
}

void te::qt::widgets::RasterEditWizardPage::setMapDisplay(te::qt::widgets::MapDisplay* mapDisplay)
{
  m_navigator->setMapDisplay(mapDisplay);
  m_mapDisplay = mapDisplay;
}

void te::qt::widgets::RasterEditWizardPage::setActionGroup(QActionGroup* actionGroup)
{
  //configure tools
  m_navigator->setActionGroup(actionGroup);
  m_navigator->enableGeomAction();
  m_navigator->enableBoxAction();
}

te::map::AbstractLayerPtr te::qt::widgets::RasterEditWizardPage::getLayer()
{
  return m_layer;
}

te::qt::widgets::RasterEditWizardPage::RASTER_EDITION_TYPE
  te::qt::widgets::RasterEditWizardPage::getEditionType()
{
  int index = m_ui->m_editionTypeComboBox->currentIndex();
  return (RASTER_EDITION_TYPE)m_ui->m_editionTypeComboBox->itemData(index).toInt();
}

int te::qt::widgets::RasterEditWizardPage::getBand()
{
  if( m_ui->m_bandComboBox->count() )
  {
    return m_ui->m_bandComboBox->currentText().toInt();
  }
  else
  {
    return 0;
  }
}

void te::qt::widgets::RasterEditWizardPage::getROIs(std::vector<te::gm::Geometry*> &geomVec)
{
  geomVec = m_geomVec;
}

void te::qt::widgets::RasterEditWizardPage::addRoiItem()
{
  int row = m_ui->m_roiTableWidget->rowCount();

  m_ui->m_roiTableWidget->insertRow(row);

  QTableWidgetItem* item = new QTableWidgetItem(
    "ROI Item " + QString::number( row + 1 ) );
  m_ui->m_roiTableWidget->setItem(row, 0, item);

  m_ui->m_roiTableWidget->resizeRowsToContents();
  m_ui->m_roiTableWidget->resizeColumnToContents(0);
}

void te::qt::widgets::RasterEditWizardPage::addValuesPushButtonClicked()
{
  double oldValue = m_ui->m_oldValueSpinBox->value();
  double newValue = m_ui->m_newValueSpinBox->value();
  int rowsCount = m_ui->m_roiTableWidget->rowCount();

  m_ui->m_roiSubstitutionValuesTableWidget->insertRow( rowsCount );

  QTableWidgetItem* oldItem = new QTableWidgetItem( QString::number(oldValue));
  m_ui->m_roiSubstitutionValuesTableWidget->setItem(rowsCount, 0, oldItem);

  QTableWidgetItem* newItem = new QTableWidgetItem( QString::number(newValue));
  m_ui->m_roiSubstitutionValuesTableWidget->setItem(rowsCount, 1, newItem);

  m_ui->m_roiSubstitutionValuesTableWidget->resizeRowsToContents();
  m_ui->m_roiSubstitutionValuesTableWidget->resizeColumnToContents(0);

  emit completeChanged();
}

void te::qt::widgets::RasterEditWizardPage::removeValuesPushButtonClicked()
{
  QList<QTableWidgetItem *> selectedItens =
    m_ui->m_roiSubstitutionValuesTableWidget->selectedItems();

  while( !selectedItens.empty() )
  {
    QTableWidgetItem * itemPtr = selectedItens.front();
    selectedItens.pop_front();
    int itemRow = itemPtr->row();
    delete itemPtr;

    m_ui->m_roiSubstitutionValuesTableWidget->removeRow( itemRow );

    selectedItens =
      m_ui->m_roiSubstitutionValuesTableWidget->selectedItems();
  }

  emit completeChanged();
}

void te::qt::widgets::RasterEditWizardPage::onEnvelopeAcquired(te::gm::Envelope env)
{
  if( m_layer )
  {
    if(m_mapDisplay->getSRID() != m_layer->getSRID())
      env.transform(m_mapDisplay->getSRID(), m_layer->getSRID());

    if (!env.isValid() || !env.intersects(m_layer->getExtent()))
      return;

    m_geomVec.push_back(te::gm::GetGeomFromEnvelope(&env, m_layer->getSRID()));

    addRoiItem();

    drawGeom();

    emit completeChanged();
  }
}

void te::qt::widgets::RasterEditWizardPage::onGeomAquired(te::gm::Polygon *poly)
{
  if( m_layer )
  {
    if(m_mapDisplay->getSRID() != m_layer->getSRID())
      poly->transform(m_layer->getSRID());

    if(!poly->isValid() || !poly->intersects(te::gm::GetGeomFromEnvelope(&m_layer->getExtent(), m_layer->getSRID())))
       return;

    m_geomVec.push_back(poly);

    addRoiItem();

    drawGeom();

    emit completeChanged();
  }
}

void te::qt::widgets::RasterEditWizardPage::onSelectRoiItem()
{
  drawGeom();

  int selectedRow = m_ui->m_roiTableWidget->currentIndex().row();

  const te::gm::Envelope& mapExt = m_mapDisplay->getExtent();

  te::qt::widgets::Canvas canvasInstance(m_mapDisplay->getDraftPixmap());
  canvasInstance.setWindow(mapExt.m_llx, mapExt.m_lly, mapExt.m_urx, mapExt.m_ury);

  canvasInstance.setPolygonContourWidth(1);
  canvasInstance.setPolygonContourColor(te::color::RGBAColor(100, 177, 216, 222));
  canvasInstance.setPolygonFillColor(te::color::RGBAColor(100, 177, 216, 220));

  canvasInstance.draw(m_geomVec.at(selectedRow));

  m_mapDisplay->repaint();
}

void te::qt::widgets::RasterEditWizardPage::onRemoveRoiItem()
{
  int row = m_ui->m_roiTableWidget->currentIndex().row();

  if(m_geomVec.empty())
  {
    return;
  }

  m_ui->m_roiTableWidget->removeRow(row);

  m_geomVec.erase(m_geomVec.begin()+row);

  drawGeom();

  emit completeChanged();
}

void te::qt::widgets::RasterEditWizardPage::drawGeom()
{
  m_mapDisplay->getDraftPixmap()->fill(Qt::transparent);

  const te::gm::Envelope& mapExt = m_mapDisplay->getExtent();

  te::qt::widgets::Canvas canvasInstance(m_mapDisplay->getDraftPixmap());
  canvasInstance.setWindow(mapExt.m_llx, mapExt.m_lly, mapExt.m_urx, mapExt.m_ury);

  canvasInstance.setPolygonContourWidth(2);
  canvasInstance.setPolygonContourColor(te::color::RGBAColor(100, 177, 216, TE_OPAQUE));
  canvasInstance.setPolygonFillColor(te::color::RGBAColor(100, 177, 216, 80));

  te::gm::Geometry* geom = nullptr;

  int index = m_ui->m_editionTypeComboBox->currentIndex();
  int type = m_ui->m_editionTypeComboBox->itemData(index).toInt();

  if(type == EDITION_ROI_ASSISTED)
  {
    const std::size_t geomVecSize = m_geomVec.size();

    for(std::size_t i = 0; i < geomVecSize; ++i)
    {
      geom = m_geomVec[i];

      if( geom->getSRID() != m_mapDisplay->getSRID())
      {
        geom->transform(m_mapDisplay->getSRID());
      }

      canvasInstance.draw( geom );
    }
  }

  m_mapDisplay->repaint();
}

void te::qt::widgets::RasterEditWizardPage::getReplaceValues(
  std::vector< std::pair< double, double > >& replaceValues ) const
{
  replaceValues.clear();

  const int rowsCount = m_ui->m_roiSubstitutionValuesTableWidget->rowCount();
  std::pair< double, double > auxPair;
  QTableWidgetItem const* itemPtr = 0;

  for( int row = 0 ; row < rowsCount ; ++row )
  {
    itemPtr = m_ui->m_roiSubstitutionValuesTableWidget->item( row, 0 );
    auxPair.first = itemPtr->text().toDouble();

    itemPtr = m_ui->m_roiSubstitutionValuesTableWidget->item( row, 1 );
    auxPair.second = itemPtr->text().toDouble();

    replaceValues.push_back( auxPair );
  }
}


void te::qt::widgets::RasterEditWizardPage::clearCanvas()
{
  te::qt::widgets::Canvas canvasInstance(m_mapDisplay->getDraftPixmap());

  canvasInstance.clear();

  m_mapDisplay->repaint();
}
