/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

/*!
  \file terralib/raster/Utils.cpp

  \brief Utility functions for the raster module.
*/

// TerraLib
#include "../raster/RasterSynchronizer.h"
#include "../raster/SynchronizedRaster.h"
#include "../datatype/Enums.h"
#include "../geometry/Coord2D.h"
#include "../common/MathUtils.h"
#include "../common/PlatformUtils.h"
#include "BandIterator.h"
#include "Exception.h"
#include "PositionIterator.h"
#include "RasterFactory.h"
#include "Utils.h"

#ifdef TERRALIB_LEAK_DETECTOR_ENABLED
  #include <vld.h>
#endif

// Boost
#include <boost/cstdint.hpp>
#include <boost/random.hpp>
#include <boost/random/uniform_int_distribution.hpp>

// STL
#include <cassert>
#include <limits>
#include <vector>
#include <thread>
#include <mutex>

static int sg_pixelSize[] = {  0,
                               0,
                               0,
                               sizeof(char) * 8,
                               sizeof(unsigned char),
                               sizeof(boost::int16_t),
                               sizeof(boost::uint16_t),
                               sizeof(boost::int32_t),
                               sizeof(boost::uint32_t),
                               sizeof(boost::int64_t),
                               sizeof(boost::uint64_t),
                               sizeof(bool),
                               sizeof(float),
                               sizeof(double),
                               0,
                               0,
                               0,
                               0,
                               0,
                               0,
                               0,
                               0,
                               0,
                               sizeof(boost::int16_t) * 2,
                               sizeof(boost::int32_t) * 2,
                               sizeof(float) * 2,
                               sizeof(double) * 2,
                               0,
                               0,
                               0,
                               sizeof(unsigned char), // 4bits = 4/8 sizeof(unsigned char)
                               sizeof(unsigned char), // 2bits = 2/8 sizeof(unsigned char)
                               sizeof(unsigned char)  // 1bit = 1/8 sizeof(unsigned char)
                            };

namespace te
{
  namespace rst
  {
    struct GetHistogramsThreadParameters
    {
      te::rst::RasterSynchronizer* m_rasterSyncPtr;
      unsigned int m_inputBandIndex;

      // position:             0:proc 1:block X     2:block Y     3:minr  4:maxr  5:mini  6:maxi
      std::vector< std::tuple< bool,  unsigned int, unsigned int, double, double, double, double > >* m_rasterBlocksStatusPtr;

      unsigned int m_bandIndex;
      unsigned int m_rowStart;
      unsigned int m_colStart;
      unsigned int m_finalRow;
      unsigned int m_finalCol;
      unsigned int m_histoBins;
      unsigned int m_sampleStep;
      bool m_returnStatus;
      std::mutex* m_mutexPtr;
      std::map<double, unsigned>* m_rHistogramPtr;
      std::map<double, unsigned>* m_iHistogramPtr;
      bool m_customHistogramBins;

      // 0 = calc min and max for each block
      // 1 = calc histograms
      unsigned int m_threadExecMode;
    };
  }
}

int te::rst::GetPixelSize(int datatype)
{
  return sg_pixelSize[datatype];
}

std::vector<te::rst::BandProperty*> te::rst::GetBandProperties(const std::map<std::string, std::string>& rinfo)
{
  std::size_t nBands;
  int tBands;
  std::vector<te::rst::BandProperty*> bprops;

  std::map<std::string, std::string>::const_iterator it = rinfo.find("NBANDS");

  if(it == rinfo.end())
    throw Exception(("Number of bands must be set!"));

  nBands = atoi(it->second.c_str());

  it = rinfo.find("BANDSTYPE");

  if(it == rinfo.end())
    throw Exception(("Data type of bands must be set!"));

  tBands = atoi(it->second.c_str());

  for (std::size_t b = 0; b < nBands; b++)
    bprops.push_back(new te::rst::BandProperty(b, tBands));

  return bprops;
}

te::rst::Grid* te::rst::GetGrid(const std::map<std::string, std::string>& rinfo)
{
  unsigned nCols, nRows;
  double resX = 1.0, resY = 1.0, ulX =0.0, ulY=0.0;
  int srid = TE_UNKNOWN_SRS;
  te::gm::Coord2D* ulc;

  std::map<std::string, std::string>::const_iterator it = rinfo.find("NCOLS");

  if(it == rinfo.end())
    throw Exception(("Number of columns must be set!"));

  nCols = atoi(it->second.c_str());

  it = rinfo.find("NROWS");

  if(it == rinfo.end())
    throw Exception(("Number of rows must be set!"));

  nRows = atoi(it->second.c_str());

  it = rinfo.find("RESX");
  if(it != rinfo.end())
    resX = atof(it->second.c_str());

  it = rinfo.find("RESY");
  if(it != rinfo.end())
    resY = atof(it->second.c_str());

  it = rinfo.find("SRID");
  if(it != rinfo.end())
    srid = atoi(it->second.c_str());

  it = rinfo.find("ULX");
  if(it != rinfo.end())
    ulX = atof(it->second.c_str());

  it = rinfo.find("ULY");
  if(it != rinfo.end())
    ulY = atof(it->second.c_str());

  ulc = new te::gm::Coord2D(ulX, ulY);

  return (new te::rst::Grid(nCols, nRows, resX, resY, ulc, srid));
}

void te::rst::Copy(const te::rst::Raster& rin, te::rst::Raster& rout)
{
  assert(rin.getNumberOfBands() == rout.getNumberOfBands());
  assert(rout.getAccessPolicy() == te::common::RWAccess || rout.getAccessPolicy() == te::common::WAccess);

  const std::size_t nbands = rin.getNumberOfBands();
  const unsigned int nRows = rin.getNumberOfRows();
  const unsigned int nCols = rin.getNumberOfColumns();
  unsigned int row = 0;
  unsigned int col = 0;
  std::complex< double > value;

  for(std::size_t b = 0; b < nbands; b++)
  {
    if(rin.getBand(b)->getProperty()->getType() == rout.getBand(b)->getProperty()->getType())
    {
      Copy(*rin.getBand(b), *rout.getBand(b));
    }
    else
    {
      const te::rst::Band& bin = *rin.getBand(b);
      te::rst::Band& bout = *rout.getBand(b);

      for(row = 0 ; row < nRows ; ++row)
        for(col = 0 ; col < nCols ; ++col)
        {
          bin.getValue(col, row, value);

          bout.setValue(col, row, value);
        }
    }
  }
}

void te::rst::Copy(const te::rst::Band& bin, te::rst::Band& bout)
{
  assert(*bin.getRaster()->getGrid() == *bout.getRaster()->getGrid());

// when both bands have the same data type
  if (bin.getProperty()->getType() == bout.getProperty()->getType())
  {
    unsigned char* buffer = new unsigned char[bin.getBlockSize()];

    int nblocksx = bin.getProperty()->m_nblocksx;
    int nblocksy = bin.getProperty()->m_nblocksy;

    const int blkw = bin.getProperty()->m_blkw;
    const int blkh = bin.getProperty()->m_blkh;

  // when both rasters have the same block size, copy the entire blocks from in to out
    if((blkw == bout.getProperty()->m_blkw) &&
       (blkh == bout.getProperty()->m_blkh))
    {
      for(int y = 0; y < nblocksy; ++y)
      {
        for(int x = 0; x < nblocksx; ++x)
        {
          bin.read(x, y, buffer);
          bout.write(x, y, buffer);
        }
      }
    }
  // get all values from input block, and copy pixel by pixel to the output band
    else
    {
      std::complex<double> value;

      const unsigned int ncols = bin.getRaster()->getNumberOfColumns();
      const unsigned int nrows = bin.getRaster()->getNumberOfRows();

      for(int y = 0; y < nblocksy; ++y)
      {
        for(int x = 0; x < nblocksx; ++x)
        {
          unsigned int w = blkw * (x + 1);

          w = w > ncols ? ncols : w;

          unsigned int h = blkh * (y + 1);

          h = h > nrows ? nrows : h;

          for(int r = blkh * y; r < (int)h; ++r)
          {
            for(int c = blkw * x; c < (int)w; ++c)
            {
              bin.getValue(c, r, value);

              bout.setValue(c, r, value);
            }
          }
        }
      }
    }

    delete [] buffer;
  }
// if bands have different data types, use iterator
  else
  {
    te::rst::ConstBandIterator<double> rit = te::rst::ConstBandIterator<double>::begin(&bin);

    te::rst::ConstBandIterator<double> ritend = te::rst::ConstBandIterator<double>::end(&bin);

    while(rit != ritend)
    {
      bout.setValue(rit.getColumn(), rit.getRow(), *rit);

      ++rit;
    }
  }
}

void te::rst::Copy(unsigned int drow, unsigned int dcolumn, unsigned int height, unsigned int width, const Raster& rin, Raster& rout)
{
  assert(drow + height <= rin.getNumberOfRows());
  assert(dcolumn + width <= rin.getNumberOfColumns());

// define variables for interpolation
  std::vector<std::complex<double> > v;

  for (unsigned r = drow; r < drow + height; r++)
  {
    for (unsigned c = dcolumn; c < dcolumn + width; c++)
    {
      te::gm::Coord2D inputGeo = rin.getGrid()->gridToGeo(c, r);

      te::gm::Coord2D outputGrid = rout.getGrid()->geoToGrid(inputGeo.x, inputGeo.y);

      int x = Round(outputGrid.x);
      int y = Round(outputGrid.y);

      if((x >=0 && x < (int)rout.getNumberOfColumns()) && (y >=0 && y < (int)rout.getNumberOfRows()))
      {
        rin.getValues(c, r, v);
        rout.setValues(x, y, v);
      }
    }
  }
}

te::rst::RasterPtr te::rst::CreateCopy(const te::rst::Raster& rin,
                                       const std::string& uri,
                                       const std::string &rType)
{
  std::map<std::string, std::string> rasterInfo;
  rasterInfo["URI"] = uri;

  std::vector<te::rst::BandProperty*> bandsProperties;

  unsigned int bandIndex = 0;

  for(bandIndex = 0; bandIndex < rin.getNumberOfBands(); ++bandIndex)
  {
    bandsProperties.push_back(new te::rst::BandProperty(*(rin.getBand(bandIndex )->getProperty())));
  }

  te::rst::RasterPtr outRasterPtr;

  outRasterPtr.reset( te::rst::RasterFactory::make( rType,  new te::rst::Grid(*(rin.getGrid())), bandsProperties, rasterInfo, nullptr, nullptr));

  if(outRasterPtr.get() == nullptr)
    return outRasterPtr;

  Copy(rin, *outRasterPtr);

  return outRasterPtr;
}

void te::rst::GetDataTypeRanges( const int& dataType, double& min, double& max )
{
  switch( dataType )
  {
    case te::dt::R4BITS_TYPE:
      min = 0;
      max = 15;
    break;

    case te::dt::R2BITS_TYPE:
      min = 0;
      max = 3;
    break;

    case te::dt::R1BIT_TYPE:
      min = 0;
      max = 1;
    break;

    case te::dt::UCHAR_TYPE:
      min = 0;
      max = 255;
    break;

    case te::dt::CHAR_TYPE:
      min = -127;
      max = 127;
    break;

    case te::dt::UINT16_TYPE:
      min = 0;
      max = (double)std::numeric_limits<unsigned short int>::max();
    break;

    case te::dt::INT16_TYPE:
      min = (double)std::numeric_limits<short int>::min();
      max = (double)std::numeric_limits<short int>::max();
    break;

    case te::dt::UINT32_TYPE:
      min = 0;
      max = (double)std::numeric_limits<unsigned int>::max();
    break;

    case te::dt::INT32_TYPE:
      min = (double)std::numeric_limits<int>::min();
      max = (double)std::numeric_limits<int>::max();
    break;

    case te::dt::FLOAT_TYPE:
      min = -(double)std::numeric_limits< float >::max();
      max = (double)std::numeric_limits< float >::max();
    break;

    case te::dt::DOUBLE_TYPE:
      min = -std::numeric_limits< double >::max();
      max = std::numeric_limits< double >::max();
    break;

    case te::dt::CINT16_TYPE:
      min = (double)std::numeric_limits<short int>::min();
      max = (double)std::numeric_limits<short int>::max();
    break;

    case te::dt::CINT32_TYPE:
      min = (double)std::numeric_limits<int>::min();
      max = (double)std::numeric_limits<int>::max();
    break;

    case te::dt::CFLOAT_TYPE:
      min = -(double)std::numeric_limits< float >::max();
      max = (double)std::numeric_limits< float >::max();
    break;

    case te::dt::CDOUBLE_TYPE:
      min = -std::numeric_limits< double >::max();
      max = std::numeric_limits< double >::max();
    break;

    default:
      throw te::rst::Exception("Invalid data type");
  }
}

void te::rst::FillRaster(te::rst::Raster* rin, const std::complex<double>& value)
{
  for (unsigned int b = 0; b < rin->getNumberOfBands(); b++)
    te::rst::FillBand(rin->getBand(b), value);
}

void te::rst::FillBand(te::rst::Band* bin, const std::complex<double>& value)
{
  for (unsigned int r = 0; r < bin->getRaster()->getNumberOfRows(); r++)
    for (unsigned int c = 0; c < bin->getRaster()->getNumberOfColumns(); c++)
      bin->setValue(c, r, value);
}

te::rst::RasterPtr te::rst::CropRaster(const te::rst::Raster& rin, const te::gm::Polygon& pin, const std::map<std::string, std::string>& rinfo, const std::string& rType)
{
  std::vector< te::gm::Geometry const *> geometries;
  geometries.push_back( &pin );

  return te::rst::RasterPtr( CropRaster( rin, geometries, rinfo, rType ).release() );
}

std::unique_ptr< te::rst::Raster > te::rst::CropRaster(const te::rst::Raster& rin,
  const std::vector< te::gm::Geometry const *> geometries,
  const std::map<std::string, std::string>& rinfo,
  const std::string& rType)
{
  std::unique_ptr< te::rst::Raster > outRasterPtr;

  // Guessing the output envelope
  // Checking input geometries

  te::gm::Envelope outEnvelope;

  {
    outEnvelope.m_llx = outEnvelope.m_lly = std::numeric_limits< double >::max();
    outEnvelope.m_urx = outEnvelope.m_ury = -1.0 * std::numeric_limits< double >::max();

    const std::size_t geometriesSize = geometries.size();
    std::size_t geometriesIdx = 0;
    te::gm::Geometry const * gPtr = nullptr;
    te::gm::Envelope const * gEnvelopePtr = nullptr;

    for( geometriesIdx = 0 ; geometriesIdx < geometriesSize ; ++geometriesIdx )
    {
      gPtr = geometries[ geometriesIdx ];
      if(gPtr == nullptr )
      {
        return outRasterPtr;
      }

      if(
          ( gPtr->getGeomTypeId() != te::gm::PolygonType )
          &&
          ( gPtr->getGeomTypeId() != te::gm::MultiPolygonType )
        )
      {
        return outRasterPtr;
      }

      gEnvelopePtr = gPtr->getMBR();

      outEnvelope.m_llx = std::min( outEnvelope.m_llx, gEnvelopePtr->m_llx );
      outEnvelope.m_lly = std::min( outEnvelope.m_lly, gEnvelopePtr->m_lly );
      outEnvelope.m_urx = std::max( outEnvelope.m_urx, gEnvelopePtr->m_urx );
      outEnvelope.m_ury = std::max( outEnvelope.m_ury, gEnvelopePtr->m_ury );
    }
  }

  // Creating the output raster

  unsigned int firstInputRow = 0;
  unsigned int firstInputCol = 0;

  {
    te::gm::Coord2D cllenv(rin.getGrid()->geoToGrid(outEnvelope.m_llx,
      outEnvelope.m_lly));
    te::gm::Coord2D curenv(rin.getGrid()->geoToGrid(outEnvelope.m_urx,
      outEnvelope.m_ury));

    firstInputRow =
      te::common::Round< double, unsigned int >(
        std::min(
          (double)( rin.getGrid()->getNumberOfRows() - 1 )
          ,
          std::max(
            0.0
            ,
            curenv.getY()
          )
        )
      );
    firstInputCol =
      te::common::Round< double, unsigned int >(
        std::min(
          (double)( rin.getGrid()->getNumberOfColumns() - 1 )
          ,
          std::max(
            0.0
            ,
            cllenv.getX()
          )
        )
      );
    const unsigned int lastInputRow =
      te::common::Round< double, unsigned int >(
        std::min(
          (double)( rin.getGrid()->getNumberOfRows() - 1 )
          ,
          std::max(
            0.0
            ,
            cllenv.getY()
          )
        )
      );
    const unsigned int lastInputCol =
      te::common::Round< double, unsigned int >(
        std::min(
          (double)( rin.getGrid()->getNumberOfColumns() - 1 )
          ,
          std::max(
            0.0
            ,
            curenv.getX()
          )
        )
      );

    if( ( lastInputRow <= firstInputRow ) || ( lastInputCol <= firstInputCol ) )
    {
      return outRasterPtr;
    }

    const unsigned int outputWidth = lastInputCol - firstInputCol + 1;
    const unsigned int outputHeight = lastInputRow - firstInputRow + 1;

    if( ( outputHeight == 0 ) || ( outputHeight == 0 ) )
    {
      return outRasterPtr;
    }

    te::gm::Coord2D ulc( rin.getGrid()->gridToGeo( ((double)firstInputCol) - 0.5,
      ((double)firstInputRow) - 0.5 ) );

    te::rst::Grid* grid = new te::rst::Grid( outputWidth, outputHeight,
      rin.getResolutionX(), rin.getResolutionY(), &ulc, rin.getSRID() );

    std::vector<te::rst::BandProperty*> bands;

    for (std::size_t b = 0; b < rin.getNumberOfBands(); b++)
    {
      bands.push_back(new te::rst::BandProperty(*(rin.getBand(b)->getProperty())));
      bands[ b ]->m_nblocksx = 1;
      bands[ b ]->m_nblocksy = outputHeight;
      bands[ b ]->m_blkw = outputWidth;
      bands[ b ]->m_blkh = 1;
    }

    outRasterPtr.reset( te::rst::RasterFactory::make(rType, grid, bands, rinfo, nullptr, nullptr) );
    if( outRasterPtr.get() == nullptr )
    {
      return outRasterPtr;
    }

    FillRaster(outRasterPtr.get(), rin.getBand(0)->getProperty()->m_noDataValue);

  }

  // Data copy

  {
    const std::size_t geometriesSize = geometries.size();
    std::size_t geometriesIdx = 0;
    std::vector< te::gm::Geometry * > singleGgeomsPtrs;
    std::size_t singleGgeomsPtrsIdx = 0;
    std::size_t singleGgeomsPtrsSize = 0;
    std::vector<std::complex<double> > values;
    unsigned int outRow = 0;
    unsigned int outCol = 0;

    for( geometriesIdx = 0 ; geometriesIdx < geometriesSize ; ++geometriesIdx )
    {
      for(auto p : singleGgeomsPtrs)
        delete p;
      singleGgeomsPtrs.clear();
      te::gm::Multi2Single( (te::gm::Geometry*)geometries[ geometriesIdx ], singleGgeomsPtrs );

      singleGgeomsPtrsSize = singleGgeomsPtrs.size();

      for( singleGgeomsPtrsIdx = 0 ; singleGgeomsPtrsIdx < singleGgeomsPtrsSize ;
        ++singleGgeomsPtrsIdx )
      {
        assert( singleGgeomsPtrs[ singleGgeomsPtrsIdx ]->getGeomTypeId() ==
          te::gm::PolygonType );

        PolygonIterator< double > it = PolygonIterator< double >::begin(
          outRasterPtr.get(), (te::gm::Polygon*)singleGgeomsPtrs[ singleGgeomsPtrsIdx ] );
        PolygonIterator< double > itEnd = PolygonIterator< double >::end(
          outRasterPtr.get(), (te::gm::Polygon*)singleGgeomsPtrs[ singleGgeomsPtrsIdx ] );

        while( it != itEnd )
        {
          outRow = it.getRow();
          outCol = it.getColumn();

          rin.getValues( outCol + firstInputCol, outRow + firstInputRow, values );
          outRasterPtr->setValues( outCol, outRow, values );

          ++it;
        }
      }
    }

    for(auto p : singleGgeomsPtrs)
      delete p;
    singleGgeomsPtrs.clear();
  }

  return outRasterPtr;
}

std::vector<te::gm::Point*> te::rst::GetRandomPointsInRaster(const te::rst::Raster& inputRaster, unsigned int numberOfPoints, bool ignoreNoDataValues)
{
  std::vector<te::gm::Point*> randomPoints;
  double randX;
  double randY;

  boost::random::mt19937 generator((boost::random::mt19937::result_type) time(nullptr));
  boost::random::uniform_int_distribution<> random_rows(0, inputRaster.getNumberOfRows() - 1);
  boost::random::uniform_int_distribution<> random_columns(0, inputRaster.getNumberOfColumns() - 1);

  for (unsigned int p = 0; p < numberOfPoints; p++)
  {
    if (ignoreNoDataValues == true)
    {
      bool isDummyValue = true;

      while (isDummyValue)
      {
        int col = random_columns(generator);
        int row = random_rows(generator);
        double value;

        for (std::size_t band = 0; band < inputRaster.getNumberOfBands(); ++band)
        {
          inputRaster.getValue(col, row, value, band);

          const te::rst::Band* curBand = inputRaster.getBand(band);

          if (value != curBand->getProperty()->m_noDataValue)
          {
            isDummyValue = false;
            break;
          }
        }

        if (isDummyValue == false)
        {
          inputRaster.getGrid()->gridToGeo(col, row, randX, randY);
          randomPoints.push_back(new te::gm::Point(randX, randY, inputRaster.getSRID()));
        }
      }
    }
    else
    {
      inputRaster.getGrid()->gridToGeo(random_columns(generator), random_rows(generator), randX, randY);
      randomPoints.push_back(new te::gm::Point(randX, randY, inputRaster.getSRID()));
    }
  }

  return randomPoints;
}

std::string te::rst::ConvertColorInterpTypeToString(const te::rst::ColorInterp& ci)
{
  if(ci == te::rst::UndefCInt)
    return TE_TR("Undefined");
  else if(ci == te::rst::GrayIdxCInt)
    return TE_TR("Gray");
  else if(ci == te::rst::PaletteIdxCInt)
    return TE_TR("Palette");
  else if(ci == te::rst::RedCInt)
    return TE_TR("Red");
  else if(ci == te::rst::GreenCInt)
    return TE_TR("Green");
  else if(ci == te::rst::BlueCInt)
    return TE_TR("Blue");
  else if(ci == te::rst::AlphaCInt)
    return TE_TR("Alpha");
  else if(ci == te::rst::HueCInt)
    return TE_TR("Hue");
  else if(ci == te::rst::SatCInt)
    return TE_TR("Saturation");
  else if(ci == te::rst::LigCInt)
    return TE_TR("Lightness");
  else if(ci == te::rst::CyanCInt)
    return TE_TR("Cyan");
  else if(ci == te::rst::MagentaCInt)
    return TE_TR("Magenta");
  else if(ci == te::rst::YellowCInt)
    return TE_TR("Yellow");
  else if(ci == te::rst::KeyCInt)
    return TE_TR("Key");
  else if(ci == te::rst::YCInt)
    return TE_TR("Y");
  else if(ci == te::rst::CbCInt)
    return TE_TR("Cb");
  else if(ci == te::rst::CrCInt)
    return TE_TR("Cr");

  return "";
}

std::string te::rst::ConvertPalleteInterpTypeToString(const te::rst::PaletteInterpretation& pi)
{
  if(pi == te::rst::UndefPalInt)
    return TE_TR("Undefined");
  else if(pi == te::rst::GrayPalInt)
    return TE_TR("Gray");
  else if(pi == te::rst::RGBPalInt)
    return TE_TR("RGB");
  else if(pi == te::rst::CMYKPalInt)
    return TE_TR("CMYK");
  else if(pi == te::rst::HSLPalInt)
    return TE_TR("HSL");

  return "";
}

namespace te
{
  namespace rst
  {
    void getHistogramsThreadEntry( te::rst::GetHistogramsThreadParameters* paramsPtr )
    {
      std::unique_lock< std::mutex > locker( *paramsPtr->m_mutexPtr );

      std::vector< std::tuple< bool,  unsigned int, unsigned int, double, double, double, double > >&
        rasterBlocksStatus = *paramsPtr->m_rasterBlocksStatusPtr;
      const std::size_t rasterBlocksStatusSize = rasterBlocksStatus.size();
      const unsigned int rowStart = paramsPtr->m_rowStart;
      const unsigned int colStart = paramsPtr->m_colStart;
      const unsigned int finalRow = paramsPtr->m_finalRow;
      const unsigned int finalCol = paramsPtr->m_finalCol;
      const unsigned int sampleStep = paramsPtr->m_sampleStep;
      const unsigned int threadExecMode = paramsPtr->m_threadExecMode;
      const bool customHistogramBins = paramsPtr->m_customHistogramBins;
      std::map<double, unsigned>& rHistogram = *paramsPtr->m_rHistogramPtr;
      std::map<double, unsigned>& iHistogram = *paramsPtr->m_iHistogramPtr;

      locker.unlock();

      const te::rst::SynchronizedRaster inputRaster( 1,
        *paramsPtr->m_rasterSyncPtr );
      const te::rst::Band& band = *inputRaster.getBand( paramsPtr->m_bandIndex );
      const double noDataValue = band.getProperty()->m_noDataValue;
      const unsigned int maxBlkWidth = band.getProperty()->m_blkw;
      const unsigned int maxBlkHeight = band.getProperty()->m_blkh;
      const unsigned int nRows = inputRaster.getNumberOfRows();
      const unsigned int nCols = inputRaster.getNumberOfColumns();

      unsigned int blkX = 0;
      unsigned int blkY = 0;
      unsigned int blkStartRow = 0;
      unsigned int blkStartCol = 0;
      unsigned int blkRowBound = 0;
      unsigned int blkColBound = 0;
      std::complex< double > complexValue = 0;
      double realValue = 0;
      double imagValue = 0;
      unsigned int row = 0;
      unsigned int col = 0;

      if( threadExecMode == 0 )
      {
        // dicovering band values limits

        double blkMinRValue = 0;
        double blkMinIValue = 0;
        double blkMaxRValue = 0;
        double blkMaxIValue = 0;

        for( std::size_t rasterBlocksStatusIdx = 0 ; rasterBlocksStatusIdx <
          rasterBlocksStatusSize ; ++rasterBlocksStatusIdx )
        {
          locker.lock();

          // looking for the next unprocessed block

          if(
              (
                std::get< 3 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] )
                ==
                std::numeric_limits< double >::max()
              )
              &&
              (
                std::get< 4 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] )
                ==
                -1.0 * std::numeric_limits< double >::max()
              )
              &&
              (
                std::get< 5 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] )
                ==
                std::numeric_limits< double >::max()
              )
              &&
              (
                std::get< 6 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] )
                ==
                -1.0 * std::numeric_limits< double >::max()
              )
            )
          {
            std::get< 3 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) = 0;
            std::get< 4 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) = 0;
            std::get< 5 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) = 0;
            std::get< 6 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) = 0;

            blkX = std::get< 1 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] );
            blkY = std::get< 2 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] );

            locker.unlock();

            // definng the current block limits

            blkStartRow = maxBlkHeight * blkY;
            blkStartRow = std::max( blkStartRow, rowStart );
            blkStartRow = std::min( blkStartRow, nRows );

            blkStartCol = maxBlkWidth * blkX;
            blkStartCol = std::max( blkStartCol, colStart );
            blkStartCol = std::min( blkStartCol, nCols );

            blkRowBound = ( maxBlkHeight * blkY ) + maxBlkHeight;
            blkRowBound = std::min( blkRowBound, finalRow + 1 );
            blkRowBound = std::min( blkRowBound, nRows );

            blkColBound = ( maxBlkWidth * blkX ) + maxBlkWidth;
            blkColBound = std::min( blkColBound, finalCol + 1 );
            blkColBound = std::min( blkColBound, nCols );

            blkMinRValue = std::numeric_limits< double >::max();
            blkMaxRValue = -1.0 * blkMinRValue;
            blkMinIValue = blkMinRValue;
            blkMaxIValue = blkMaxRValue;

            // computing block value ranges

            for( row = blkStartRow ; row < blkRowBound ; row += sampleStep )
            {
              for( col = blkStartCol ; col < blkColBound ; col += sampleStep )
              {
                band.getValue( col, row, complexValue );
                realValue = complexValue.real();
                imagValue = complexValue.imag();

                if( realValue != noDataValue )
                {
                  blkMinRValue = std::min( blkMinRValue, realValue );
                  blkMaxRValue = std::max( blkMaxRValue, realValue );

                  blkMinIValue = std::min( blkMinIValue, imagValue );
                  blkMaxIValue = std::max( blkMaxIValue, imagValue );
                }
              }
            }

            locker.lock();

            std::get< 3 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) =
              blkMinRValue;
            std::get< 4 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) =
              blkMaxRValue;
            std::get< 5 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) =
              blkMinIValue;
            std::get< 6 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) =
              blkMaxIValue;

            locker.unlock();
          }
          else
          {
            locker.unlock();
          }
        }
      }
      else // threadExecMode == 1
      {
        locker.lock();

        std::map<double, unsigned> blkRHistogram =
          *paramsPtr->m_rHistogramPtr;
        std::map<double, unsigned> blkIHistogram =
          *paramsPtr->m_iHistogramPtr;
        std::map<double, unsigned>::iterator hIt;
        std::map<double, unsigned>::iterator hItEnd;

        locker.unlock();

        // generating histograms

        for( std::size_t rasterBlocksStatusIdx = 0 ; rasterBlocksStatusIdx <
          rasterBlocksStatusSize ; ++rasterBlocksStatusIdx )
        {
          locker.lock();

          // looking for the next unprocessed block

          if(
              std::get< 0 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] )
              ==
              false
            )
          {
            std::get< 0 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) = true;

            blkX = std::get< 1 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] );
            blkY = std::get< 2 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] );

            locker.unlock();

            // clear block histograms

            if( customHistogramBins )
            {
              hIt = blkRHistogram.begin();
              hItEnd = blkRHistogram.end();
              while( hIt != hItEnd )
              {
                hIt->second = 0;
                ++hIt;
              }

              hIt = blkIHistogram.begin();
              hItEnd = blkIHistogram.end();
              while( hIt != hItEnd )
              {
                hIt->second = 0;
                ++hIt;
              }
            }
            else
            {
              blkIHistogram.clear();
              blkRHistogram.clear();
            }

            // definng the current block limits

            blkStartRow = maxBlkHeight * blkY;
            blkStartRow = std::max( blkStartRow, rowStart );
            blkStartRow = std::min( blkStartRow, nRows );

            blkStartCol = maxBlkWidth * blkX;
            blkStartCol = std::max( blkStartCol, colStart );
            blkStartCol = std::min( blkStartCol, nCols );

            blkRowBound = ( maxBlkHeight * blkY ) + maxBlkHeight;
            blkRowBound = std::min( blkRowBound, finalRow + 1 );
            blkRowBound = std::min( blkRowBound, nRows );

            blkColBound = ( maxBlkWidth * blkX ) + maxBlkWidth;
            blkColBound = std::min( blkColBound, finalCol + 1 );
            blkColBound = std::min( blkColBound, nCols );

            for( row = blkStartRow ; row < blkRowBound ; row += sampleStep )
            {
              for( col = blkStartCol ; col < blkColBound ; col += sampleStep )
              {
                band.getValue( col, row, complexValue );
                realValue = complexValue.real();
                imagValue = complexValue.imag();

                if( realValue != noDataValue )
                {
                  if( customHistogramBins )
                  {
                    ++( blkRHistogram.lower_bound( realValue )->second );
                    ++( blkIHistogram.lower_bound( imagValue )->second );
                  }
                  else
                  {
                    ++blkRHistogram[ realValue ];
                    ++blkIHistogram[ imagValue ];
                  }
                }
              }
            }

            // transfering values to the main histograms

            locker.lock();

            hIt = blkRHistogram.begin();
            hItEnd = blkRHistogram.end();
            while( hIt != hItEnd )
            {
              rHistogram[ hIt->first ] += hIt->second;
              ++hIt;
            }

            hIt = blkIHistogram.begin();
            hItEnd = blkIHistogram.end();
            while( hIt != hItEnd )
            {
              iHistogram[ hIt->first ] += hIt->second;
              ++hIt;
            }

            locker.unlock();
          }
          else
          {
            locker.unlock();
          }
        }
      }
    }
  }
}

bool te::rst::getRealHistogram(
      const te::rst::Raster& raster,
      const unsigned int bandIndex,
      std::map<double, unsigned>& rHistogram )
{
  std::map<double, unsigned> iHistogram;
  return getHistograms(
    raster,
    bandIndex,
    0,
    0,
    raster.getNumberOfRows() - 1,
    raster.getNumberOfColumns() - 1,
    0,
    1,
    0,
    rHistogram,
    iHistogram );
}


bool te::rst::getHistograms(
      const te::rst::Raster& raster,
      const unsigned int bandIndex,
      const unsigned int rowStart,
      const unsigned int colStart,
      const unsigned int finalRow,
      const unsigned int finalCol,
      const unsigned int histoBins,
      const unsigned int sampleStep,
      const unsigned int maxThreads,
      std::map<double, unsigned>& rHistogram,
      std::map<double, unsigned>& iHistogram )
{
  rHistogram.clear();
  iHistogram.clear();

  if(
      ( bandIndex >= raster.getNumberOfBands() )
      ||
      ( rowStart > finalRow )
      ||
      ( colStart > finalCol )
      ||
      ( finalRow >= raster.getNumberOfRows() )
      ||
      ( finalCol >= raster.getNumberOfColumns() )
      ||
      ( sampleStep == 0 )
    )
  {
    return false;
  }

  const te::rst::Band& band = *raster.getBand( bandIndex );

  bool customHistogramBins = false;
  if(
      ( band.getProperty()->m_type == te::dt::FLOAT_TYPE )
      ||
      ( band.getProperty()->m_type == te::dt::DOUBLE_TYPE )
      ||
      ( band.getProperty()->m_type == te::dt::CFLOAT_TYPE )
      ||
      ( band.getProperty()->m_type == te::dt::CDOUBLE_TYPE )
      ||
      ( histoBins != 0 )
    )
  {
    customHistogramBins = true;
  }
  if( customHistogramBins && ( histoBins == 0 ) ) return false;

  std::vector< std::tuple< bool,  unsigned int, unsigned int, double, double, double, double > >
    rasterBlocksStatus;

  for( unsigned int blkX = 0 ; blkX < band.getProperty()->m_nblocksx ; ++blkX )
  {
    for( unsigned int blkY = 0 ; blkY < band.getProperty()->m_nblocksy ; ++blkY )
    {
      std::tuple< bool,  unsigned int, unsigned int, double, double, double, double > tuple;
      std::get< 0 >( tuple ) = false;
      std::get< 1 >( tuple ) = blkX;
      std::get< 2 >( tuple ) = blkY;
      std::get< 3 >( tuple ) = std::numeric_limits< double >::max();
      std::get< 4 >( tuple ) = -1.0 * std::numeric_limits< double >::max();
      std::get< 5 >( tuple ) = std::numeric_limits< double >::max();
      std::get< 6 >( tuple ) = -1.0 * std::numeric_limits< double >::max();

      rasterBlocksStatus.push_back( tuple );
    }
  }

  te::rst::RasterSynchronizer rasterSync(
    *(te::rst::Raster*)(&raster), te::common::RAccess );
  std::mutex mutex;

  te::rst::GetHistogramsThreadParameters params;
  params.m_rasterSyncPtr = &rasterSync;
  params.m_inputBandIndex = bandIndex;
  params.m_rasterBlocksStatusPtr = &rasterBlocksStatus;
  params.m_bandIndex = bandIndex;
  params.m_rowStart = rowStart;
  params.m_colStart = colStart;
  params.m_finalRow = finalRow;
  params.m_finalCol = finalCol;
  params.m_histoBins = histoBins;
  params.m_sampleStep = sampleStep;
  params.m_returnStatus = true;
  params.m_mutexPtr = &mutex;
  params.m_rHistogramPtr = &rHistogram;
  params.m_iHistogramPtr = &iHistogram;
  params.m_customHistogramBins = customHistogramBins;

  unsigned int threadsNumber = maxThreads ? maxThreads :
    te::common::GetPhysProcNumber();

  // step 1 - min and max calcule

  double histoMinRValue = std::numeric_limits< double >::max();
  double histoMaxRValue = -1.0 * histoMinRValue;
  double histoMinIValue = histoMinRValue;
  double histoMaxIValue = histoMaxRValue;

  if( customHistogramBins )
  {
    params.m_threadExecMode = 0;

    if( maxThreads == 1 )
    {
      getHistogramsThreadEntry( &params );
    }
    else
    {
      std::vector< std::shared_ptr< std::thread > > threads;

      for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
        ++threadIdx )
      {
        std::shared_ptr< std::thread > threadPtr( new std::thread(
          &getHistogramsThreadEntry, &params ) );

        threads.push_back( threadPtr );
      };

      for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
        ++threadIdx )
      {
        threads[ threadIdx ]->join();
      };
    }

    if( params.m_returnStatus == false ) return false;

    const std::size_t rasterBlocksStatusSize = rasterBlocksStatus.size();


    for( std::size_t rasterBlocksStatusIdx = 0 ; rasterBlocksStatusIdx <
      rasterBlocksStatusSize ; ++rasterBlocksStatusIdx )
    {
      histoMinRValue = std::min( histoMinRValue,
        std::get< 3 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) );
      histoMaxRValue = std::max( histoMaxRValue,
        std::get< 4 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) );

      histoMinIValue = std::min( histoMinIValue,
        std::get< 5 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) );
      histoMaxIValue = std::max( histoMaxIValue,
        std::get< 6 >( rasterBlocksStatus[ rasterBlocksStatusIdx ] ) );
    }
  }

  // step 2 - histogram bins

  if( customHistogramBins )
  {
    const double binRValue =
      ( histoMaxRValue - histoMinRValue ) / (double)histoBins;
    const double binIValue =
      ( histoMaxIValue - histoMinIValue ) / (double)histoBins;

    for( unsigned int binIdx = 0 ; binIdx < histoBins ; ++binIdx )
    {
      rHistogram[ ( ((double)binIdx) * binRValue ) + histoMinRValue ] = 0;
      iHistogram[ ( ((double)binIdx) * binIValue ) + histoMinIValue ] = 0;
    }
  }

  // step 3 - histograms frequencies calcule

  params.m_threadExecMode = 2;

  if( maxThreads == 1 )
  {
    getHistogramsThreadEntry( &params );
  }
  else
  {
    std::vector< std::shared_ptr< std::thread > > threads;

    for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
      ++threadIdx )
    {
      std::shared_ptr< std::thread > threadPtr( new std::thread(
        &getHistogramsThreadEntry, &params ) );

      threads.push_back( threadPtr );
    };

    for( unsigned int threadIdx = 0 ; threadIdx < threadsNumber ;
      ++threadIdx )
    {
      threads[ threadIdx ]->join();
    };
  }

  // step 4 - fill histogram gaps

  if( ! customHistogramBins )
  {
    if( ! rHistogram.empty() )
    {
      std::map<double, unsigned int>::const_iterator it = rHistogram.begin();
      std::map<double, unsigned int>::const_reverse_iterator rit = rHistogram.rbegin();

      const double min = it->first;
      const double max = rit->first;

      for( double hValue = min ; hValue <= max ; ++hValue )
      {
        rHistogram[ hValue ] += 0;
      }
    }

    if( ! iHistogram.empty() )
    {
      std::map<double, unsigned int>::const_iterator it = iHistogram.begin();
      std::map<double, unsigned int>::const_reverse_iterator rit = iHistogram.rbegin();

      const double min = it->first;
      const double max = rit->first;

      for( double hValue = min ; hValue <= max ; ++hValue )
      {
        iHistogram[ hValue ] += 0;
      }
    }
  }

  return params.m_returnStatus;
}

void te::rst::getHistogramStats(
  const std::map<double, unsigned>& histogram,
  double& min,
  double& max,
  double& mean,
  double& stdDev )
{
  min = max = mean = stdDev = 0;

  if( ! histogram.empty() )
  {
    std::map<double, unsigned int>::const_iterator it = histogram.begin();
    std::map<double, unsigned int>::const_reverse_iterator rit = histogram.rbegin();
    const std::map<double, unsigned int>::const_iterator itEnd = histogram.end();

    min = it->first;
    max = rit->first;

    double elementsNumber = 0;
    while( it != itEnd )
    {
      mean += ( it->first * ((double)it->second) );
      elementsNumber += it->second;
      ++it;
    }
    if( elementsNumber != 0 )
    {
      mean /= elementsNumber;
    }

    double diff = 0;
    it = histogram.begin();
    while( it != itEnd )
    {
      diff = ( it->first - mean );
      stdDev += ( ( diff * diff ) * ((double)it->second) );
      ++it;
    }
    if( elementsNumber != 0 )
    {
      stdDev = std::sqrt( stdDev / elementsNumber );
    }
  }
}
