/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

#include "SubdivideByBorderTillingStrategy.h"

#include "PolygonSubdivider.h"
#include "SubdividerTilling.h"
#include "../common/StringUtils.h"
#include "../common/STLUtils.h"
#include "../common/progress/TaskProgress.h"
#include "../core/filesystem/FileSystem.h"
#include "../core/logger/Logger.h"
#include "../core/translator/Translator.h"
#include "../dataaccess/dataset/DataSetAdapter.h"
#include "../dataaccess/dataset/DataSetTypeConverter.h"
#include "../dataaccess/utils/DataAccess.h"
#include "../dataaccess/utils/Utils.h"
#include "../datatype/SimpleData.h"
#include "../geometry/CommonDataStructures.h"
#include "../geometry/GeometryFunctions.h"
#include "../geometry/GeometryProperty.h"
#include "../geometry/Utils.h"
#include "../geometry/Validation.h"
#include "../memory/DataSet.h"
#include "../memory/DataSetItem.h"
#include "../memory/PreparedItemCopy.h"
#include "../srs/Converter.h"

#include "CommonFunctions.h"
#include "Feature.h"
#include "GeometrySnapper.h"
#include "Utils.h"

 //Boost
#include <boost/thread.hpp>
#include <boost/asio/io_service.hpp>

class SubdivideByBorderTillingStrategyImpl
{
  public:
    SubdivideByBorderTillingStrategyImpl(const te::gm::Envelope& mbr, int srid, double resolutionX, double resolutionY, te::common::TaskProgress* taskProgress)
      : m_srid(srid)
      , m_tilling(mbr, resolutionX, resolutionY)
      , m_taskProgress(taskProgress)
    {

    }

    ~SubdivideByBorderTillingStrategyImpl()
    {
      for (std::size_t i = 0; i < m_vecTemporaryDataSources.size(); ++i)
      {
        te::da::DataSourcePtr dataSource = m_vecTemporaryDataSources.at(i);
        dataSource->close();
        te::vp::DropTemporaryDataSource(dataSource);
      }
    }

    te::da::DataSourcePtr createTempDataSource()
    {
      te::da::DataSourcePtr dataSource = te::vp::CreateTemporaryDataSource();
      m_vecTemporaryDataSources.push_back(dataSource);

      return dataSource;
    }

    std::vector<te::da::DataSourcePtr> m_vecTemporaryDataSources;
    int m_srid;
    te::vp::SubdividerTilling m_tilling;
    te::common::TaskProgress* m_taskProgress;
};

te::vp::SubdivideByBorderTillingStrategy::SubdivideByBorderTillingStrategy(const te::gm::Envelope& mbr, int srid, double resolutionX, double resolutionY, te::common::TaskProgress* taskProgress)
  : AbstractTillingStrategy()
  , m_impl(new SubdivideByBorderTillingStrategyImpl(mbr, srid, resolutionX, resolutionY, taskProgress))
{
  init();
}

te::vp::SubdivideByBorderTillingStrategy::~SubdivideByBorderTillingStrategy()
{
  delete m_impl;
}

void te::vp::SubdivideByBorderTillingStrategy::init()
{
}

te::da::DataAccess* CreateOutputDataAccess(const te::da::DataSetType* inputDataType, te::da::DataSourcePtr outputDataSource, const std::string& outputDataSetName, int srid)
{
  if (outputDataSource->isOpened() == false)
  {
    outputDataSource->open();
  }

  std::string lowerName = te::common::Convert2LCase(outputDataSetName);

  std::unique_ptr<te::da::DataSetType> outputDataSetType(static_cast<te::da::DataSetType*>(inputDataType->clone())); 
  outputDataSetType->setName(lowerName);
  outputDataSetType->setTitle(lowerName);
  outputDataSetType->setDatasetName(lowerName);

  te::gm::GeometryProperty* geometryProperty = te::da::GetFirstGeomProperty(outputDataSetType.get());
  geometryProperty->setGeometryType(te::gm::GetMultiType(geometryProperty->getGeometryType()));
  if (srid != TE_UNKNOWN_SRS)
  {
    geometryProperty->setSRID(srid);
  }

  te::da::DataAccess* outputDataAccess = new te::da::DataAccess(outputDataSource, outputDataSetType->getName());
  outputDataAccess->createDataSet(outputDataSetType.get());
  return outputDataAccess;
}

struct ClipResult
{
  ClipResult(std::size_t index)
    : m_index(index)
  {

  }

  ~ClipResult()
  {
    te::common::FreeContents(m_outputFeatureSet);
  }

  std::size_t m_index;
  std::unique_ptr<te::vp::Feature> m_inputFeature;
  te::vp::FeatureSet m_outputFeatureSet;
};

void ClipFeatureThread(const te::vp::SubdividerTilling& tilling, const te::da::DataSetType* dataSetType, ClipResult* clipResult, std::size_t maxCoordinates, const te::mem::PreparedItemCopy& preparedItemCopy)
{
  te::gm::GeometryProperty* geometryProperty = te::da::GetFirstGeomProperty(dataSetType);
  te::gm::GeomType expectedGeometryType = geometryProperty->getGeometryType();

  std::size_t parentIndex = clipResult->m_index;
  const te::vp::Feature* inputFeature = clipResult->m_inputFeature.get();
  std::unique_ptr<te::gm::Geometry> geometry = inputFeature->getFirstGeometry();

  //we must fix the geometry once
  geometry = te::gm::Validation::MakeValid(geometry.get());
  if (te::gm::IsNullOrEmpty(geometry.get()) == true)
  {
    return;
  }

  const te::gm::Envelope* mbr = geometry->getMBR();

  std::size_t initRow = tilling.getTileRow(mbr->getLowerLeft());
  std::size_t initColumn = tilling.getTileColumn(mbr->getLowerLeft());

  std::size_t finalRow = tilling.getTileRow(mbr->getUpperRight());
  std::size_t finalColumn = tilling.getTileColumn(mbr->getUpperRight());

  std::string columnPropertyName = te::vp::SubdivideByBorderTillingStrategy::getFullPropertyName(dataSetType->getName(), te::vp::SubdivideByBorderTillingStrategy::getColumnPropertyName());
  std::string rowPropertyName = te::vp::SubdivideByBorderTillingStrategy::getFullPropertyName(dataSetType->getName(), te::vp::SubdivideByBorderTillingStrategy::getRowPropertyName());
  std::string parentPropertyName = te::vp::SubdivideByBorderTillingStrategy::getFullPropertyName(dataSetType->getName(), te::vp::SubdivideByBorderTillingStrategy::getParentPropertyName());

  std::size_t subdivideColumnPos = dataSetType->getPropertyPosition(columnPropertyName);
  std::size_t subdivideRowPos = dataSetType->getPropertyPosition(rowPropertyName);
  std::size_t subdivideParentIndexPos = dataSetType->getPropertyPosition(parentPropertyName);

  te::vp::FeatureSet featureSet;
  for (std::size_t row = initRow; row <= finalRow; ++row)
  {
    for (std::size_t column = initColumn; column <= finalColumn; ++column)
    {
      te::gm::Envelope clipArea = tilling.getTile(column, row);
      te::gm::GeometryVector vecClipped = te::vp::PolygonSubdivider::subdivide(geometry.get(), clipArea, maxCoordinates);

      for (std::size_t f = 0; f < vecClipped.size(); ++f)
      {
        te::gm::GeometryVector vecAdjusted = te::vp::AdjustToGeomType(vecClipped.at(f), expectedGeometryType);
        for (std::size_t j = 0; j < vecAdjusted.size(); ++j)
        {
          //we now clone the attributes
          std::unique_ptr<te::vp::Feature> featureFragment(new te::vp::Feature(dataSetType));
          preparedItemCopy.copyValues(inputFeature, featureFragment.get(), false);
          featureFragment->setFirstGeometry(vecAdjusted.at(j));
          featureFragment->setInt32(subdivideColumnPos, static_cast<boost::int32_t>(column));
          featureFragment->setInt32(subdivideRowPos, static_cast<boost::int32_t>(row));
          featureFragment->setInt32(subdivideParentIndexPos, static_cast<boost::int32_t>(parentIndex));

          featureSet.push_back(featureFragment.release());
        }
      }

      te::common::FreeContents(vecClipped);
    }
  }

  clipResult->m_outputFeatureSet = featureSet;
}
boost::mutex m_mutex;

void MultiClipFeatureThread(const te::vp::SubdividerTilling& tilling, te::da::DataAccess* dataAccess, te::srs::Converter* converter, const std::vector<ClipResult*>& vecClipResult, std::size_t maxCoordinates, te::common::TaskProgress* taskProgress)
{
  const te::da::DataSetType* dataSetType = dataAccess->getDataSetType();
  std::unique_ptr<te::mem::DataSet> outputDataSet(new te::mem::DataSet(dataSetType));

  //this object caches the column names to make faster the copy of attributes
  te::mem::PreparedItemCopy preparedItemCopy;
  for (std::size_t i = 0; i < vecClipResult.size(); ++i)
  {
    ClipResult* clipResult = vecClipResult.at(i);

    //we ensure that the feature is in the tilling reference projection
    if (converter != nullptr)
    {
      clipResult->m_inputFeature->transform(converter);
    }

    ClipFeatureThread(tilling, dataSetType, clipResult, maxCoordinates, preparedItemCopy);

    if (taskProgress != nullptr)
    {
      taskProgress->pulse();
    }
  }

  for (std::size_t i = 0; i < vecClipResult.size(); ++i)
  {
    const te::vp::FeatureSet& outputFeatureSet = vecClipResult.at(i)->m_outputFeatureSet;
    for (std::size_t j = 0; j < outputFeatureSet.size(); ++j)
    {
      te::vp::Feature* outputFeature = outputFeatureSet.at(j);
      outputDataSet->add(outputFeature);
    }

    //we must clear the list here to avoid double free at the end of this function
    vecClipResult.at(i)->m_outputFeatureSet.clear();
  }

  te::common::FreeContents(vecClipResult);

  //the lock is only used in the saving of the geometries
  boost::lock_guard<boost::mutex> lockGuard(m_mutex);

  //this function also deletes the items referenced from 'vecFullClipResult'
  try
  {
    dataAccess->saveDataSet(outputDataSet.release(), false);
  }
  catch (const std::exception& e)
  {
    std::string logMessage = "Error saving subdivided data: ";
    logMessage += e.what();
    TE_LOG_ERROR(logMessage);
  }
}

te::da::DataAccess* te::vp::SubdivideByBorderTillingStrategy::subdivide(const te::da::DataAccess* inputDataAccess, std::size_t maxCoordinates)
{
  //now we must copy the data from the input layers to the sqlite database, clipping all the geometries using the subdivider tilling
  std::string outputDataSetName = inputDataAccess->getDataSetName() + "_subdivided";
  outputDataSetName = te::da::GetAliasName(outputDataSetName);

  te::da::DataAccess* outputDataAccess = createInputDataAccessForSubdivision(inputDataAccess->getDataSetType(), outputDataSetName);
 
  std::unique_ptr<te::da::DataSet> inputDataSet = inputDataAccess->getDataSet();
  inputDataSet->moveBeforeFirst();

  std::size_t geometryPos = te::da::GetFirstSpatialPropertyPos(inputDataSet.get());

  // we create a thread pool so we can use all the available cores
  boost::asio::io_service ioService;
  boost::thread_group threadGroup;
  boost::asio::io_service::work work(ioService);

  //the we add the threads to the pool
  std::size_t numThreads = static_cast<std::size_t>(boost::thread::hardware_concurrency());
  for (std::size_t i = 0; i < numThreads; ++i)
  {
    threadGroup.create_thread(boost::bind(&boost::asio::io_service::run, &ioService));
  }

  te::mem::PreparedItemCopy copy;

  std::unique_ptr<te::srs::Converter> converter;

  bool transform = (inputDataAccess->getSRID() != m_impl->m_srid);
  if (transform == true)
  {
    converter.reset(new te::srs::Converter());
    converter->setSourceSRID(inputDataAccess->getSRID());
    converter->setTargetSRID(m_impl->m_srid);
  }

  //we create 10 tiles for each thread based on the dataSetSize
  std::size_t dataSetSize = inputDataSet->size();
  std::size_t jobSize = static_cast<std::size_t>(std::ceil(static_cast<double>(dataSetSize) / static_cast<double>(numThreads) / static_cast<double>(10)));
  if (jobSize == 0)
  {
    jobSize = dataSetSize;
  }

  if (m_impl->m_taskProgress != nullptr)
  {
    m_impl->m_taskProgress->setMessage(TE_TR("Initializing - Subdividing data") + std::string(": ") + inputDataAccess->getDataSetName());
    m_impl->m_taskProgress->setTotalSteps(static_cast<int>(dataSetSize));
    m_impl->m_taskProgress->setCurrentStep(0);
    m_impl->m_taskProgress->useMultiThread(true);
  }

  std::vector<ClipResult*> vecPartialClipResult;
  vecPartialClipResult.reserve(jobSize);

  std::size_t index = 0;
  while (inputDataSet->moveNext() == true)
  {
    //we must check if the input geometry is null. If it is, we must skip it
    if (inputDataSet->isNull(geometryPos) == true)
    {
      continue;
    }

    std::unique_ptr<te::vp::Feature> inputFeature(te::vp::Feature::createFeatureFromCurrent(inputDataSet.get()));

    ClipResult* clipResult = new ClipResult(index);
    clipResult->m_inputFeature = std::move(inputFeature);

    ++index;

    vecPartialClipResult.push_back(clipResult);
    if (vecPartialClipResult.size() == jobSize)
    {
      //adds the jobs to the manager
      ioService.post(boost::bind(MultiClipFeatureThread, m_impl->m_tilling, outputDataAccess, converter.get(), vecPartialClipResult, maxCoordinates, m_impl->m_taskProgress));

      //no thread. This code is used to run without using multithread. Useful for debug
      //MultiClipFeatureThread(m_impl->m_tilling, outputDataAccess, converter.get(), vecPartialClipResult, maxCoordinates, m_impl->m_taskProgress);

      vecPartialClipResult = std::vector<ClipResult*>();
      vecPartialClipResult.reserve(jobSize);
    }
  }

  if (vecPartialClipResult.empty() == false)
  {
    //adds the jobs to the manager
    ioService.post(boost::bind(MultiClipFeatureThread, m_impl->m_tilling, outputDataAccess, converter.get(), vecPartialClipResult, maxCoordinates, m_impl->m_taskProgress));

    //no thread. This code is used to run without using multithread. Useful for debug
    //MultiClipFeatureThread(m_impl->m_tilling, outputDataAccess, converter.get(), vecPartialClipResult, maxCoordinates, m_impl->m_taskProgress);

    vecPartialClipResult = std::vector<ClipResult*>();
  }

  //wait for the jobs to be finished
  ioService.stop();
  threadGroup.join_all();

  return outputDataAccess;
}

std::vector<te::da::DataAccess*> te::vp::SubdivideByBorderTillingStrategy::subdivide(const std::vector<te::da::DataAccess*>& vecInputDataAccess, std::size_t maxCoordinates)
{
  std::vector<te::da::DataAccess*> vecOutputDataAccess;
  for (std::size_t i = 0; i < vecInputDataAccess.size(); ++i)
  {
    //now we must copy the data from the input layers to the sqlite database, clipping all the geometries using the subdivider tilling
    te::da::DataAccess* inputDataAccess = vecInputDataAccess.at(i);

    te::da::DataAccess* outputDataAccess = subdivide(inputDataAccess, maxCoordinates);
    vecOutputDataAccess.push_back(outputDataAccess);
  }

  return vecOutputDataAccess;
}

te::vp::FeatureSet DissolveGroupMultiThread(const te::da::DataSetType* dataSetType, const te::vp::FeatureSet& fs, te::gm::GeomType expectedGeomType)
{
  te::vp::FeatureSet outputFeatureSet;
  if (fs.empty() == true)
  {
    return outputFeatureSet;
  }

  te::mem::PreparedItemCopy copy;
  if (fs.size() == 1)
  {
    const te::vp::Feature* seedFeature = fs.at(0);

    std::unique_ptr<te::vp::Feature> outputFeature(new te::vp::Feature(dataSetType));
    copy.copyValues(seedFeature, outputFeature.get(), true);

    outputFeatureSet.push_back(outputFeature.release());

    return outputFeatureSet;
  }

  const te::vp::Feature* seedFeature = fs.at(0);

  //we first snap all the features to ensure that we wont generate non representative artefacts
  te::vp::FeatureSet fsSnapped = te::vp::GeometrySnapper::snapFeatures(fs, te::gm::GetAppropriatePrecision(seedFeature->getFirstGeometryPtr()->getSRID()), true);
  if (fsSnapped.empty() == true)
  {
    return outputFeatureSet;
  }

  //then we get all the geometries from the features
  te::gm::GeometryVector vecGeometries;
  vecGeometries.reserve(fsSnapped.size());
  for (std::size_t i = 0; i < fsSnapped.size(); ++i)
  {
    const te::vp::Feature* feature = fsSnapped.at(i);
    vecGeometries.push_back(feature->getFirstGeometry().release());
  }

  te::common::FreeContents(fsSnapped);

  //the next step is to dissolve all the border from this geometry list
  te::gm::GeometryPtr result = te::gm::CascadedUnion(vecGeometries);

  te::common::FreeContents(vecGeometries);

  //then we split the possible multi geometries into single parts
  te::gm::GeometryVector vecSingle;
  te::gm::Multi2Single(result.get(), vecSingle);
  
  for(std::size_t i = 0; i < vecSingle.size(); ++i)
  {
    //we adjust the result type for it to be compatible to the expected output
    te::gm::GeometryVector vecAdjusted = te::vp::AdjustToGeomType(vecSingle.at(i), expectedGeomType);

    //and finally we create the output feature and add it to the outputFeatureSet
    for (std::size_t j = 0; j < vecAdjusted.size(); ++j)
    {
      std::unique_ptr<te::vp::Feature> outputFeature(new te::vp::Feature(dataSetType));
      copy.copyValues(seedFeature, outputFeature.get(), false);
      outputFeature->setFirstGeometry(vecAdjusted.at(j));

      outputFeatureSet.push_back(outputFeature.release());
    }
  }

  te::common::FreeContents(vecSingle);
  
  return outputFeatureSet;
}

void DissolveMultiThread(te::da::DataAccess* outputDataAccess, const std::vector<te::vp::FeatureSet>& vecGroups, te::common::TaskProgress* taskProgress)
{
  const te::da::DataSetType* outputDataSetType = outputDataAccess->getDataSetType();
  te::gm::GeometryProperty* geometryProperty = te::da::GetFirstGeomProperty(outputDataSetType);
  te::gm::GeomType outputGeomType = geometryProperty->getGeometryType();

  std::unique_ptr<te::mem::DataSet> outputDataSet(new te::mem::DataSet(outputDataSetType));

  std::size_t numProcessedGeometries = 0;
  for (std::size_t i = 0; i < vecGroups.size(); ++i)
  {
    te::vp::FeatureSet inputFeatureSet = vecGroups.at(i);
    te::vp::FeatureSet outputFeatureSet = DissolveGroupMultiThread(outputDataSetType, inputFeatureSet, outputGeomType);

    numProcessedGeometries += inputFeatureSet.size();

    te::common::FreeContents(inputFeatureSet);

    for (std::size_t j = 0; j < outputFeatureSet.size(); ++j)
    {
      outputDataSet->add(outputFeatureSet.at(j));
    }
  }

  if (outputDataSet->isEmpty() == true)
  {
    return;
  }

  //the lock is only used in the saving of the geometries
  boost::lock_guard<boost::mutex> lockGuard(m_mutex);

  //this function also deletes the items referenced from 'vecFullClipResult'
  try
  {

    if (taskProgress != nullptr)
    {
      int currentStep = taskProgress->getCurrentStep();
      taskProgress->setCurrentStep(currentStep + static_cast<int>(numProcessedGeometries));
    }
    
    outputDataAccess->saveDataSet(outputDataSet.release(), false);
  }
  catch (const std::exception & e)
  {
    std::string logMessage = "Error saving result dissolved data: ";
    logMessage += e.what();
    TE_LOG_ERROR(logMessage);
  }
}

te::da::DataAccess* te::vp::SubdivideByBorderTillingStrategy::merge(te::da::DataAccess* inputDataAccess)
{
  const te::da::DataSetType* inputDataSetType = inputDataAccess->getDataSetType();
  std::unique_ptr<te::da::DataSetType> filteredDataSetType(static_cast<te::da::DataSetType*>(inputDataSetType->clone()));

  std::vector<std::string> vecParentProperties;
  std::string parentProperties;
  for (std::size_t i = 0; i < inputDataSetType->getProperties().size(); ++i)
  {
    std::string propertyName = inputDataSetType->getProperty(i)->getName();
    if (propertyName.find(this->getColumnPropertyName()) != std::string::npos)
    {
      filteredDataSetType->remove(filteredDataSetType->getProperty(propertyName));
    }
    else if (propertyName.find(this->getRowPropertyName()) != std::string::npos)
    {
      filteredDataSetType->remove(filteredDataSetType->getProperty(propertyName));
    }
    else if (propertyName.find(this->getParentPropertyName()) != std::string::npos)
    {
      filteredDataSetType->remove(filteredDataSetType->getProperty(propertyName));
      if (parentProperties.empty() == false)
      {
        parentProperties += ",";
      }
      parentProperties += propertyName;
      vecParentProperties.push_back(propertyName);
    }
  }

  //now we must copy the data from the input layers to the sqlite database, clipping all the geometries using the subdivider tilling
  std::string outputDataSetName = inputDataAccess->getDataSetName() + "_merged";
  outputDataSetName = te::da::GetAliasName(outputDataSetName);

  te::da::DataAccess* outputDataAccess = CreateOutputDataAccess(filteredDataSetType.get(), m_impl->createTempDataSource(), outputDataSetName, m_impl->m_srid);


  const te::da::DataSetType* outputDataSetType = outputDataAccess->getDataSetType();

  te::gm::GeometryProperty* geometryProperty = te::da::GetFirstGeomProperty(outputDataSetType);
  int srid = geometryProperty->getSRID();
  te::gm::GeomType outputGeomType = geometryProperty->getGeometryType();

  // we create a thread pool so we can use all the available cores
  boost::asio::io_service ioService;
  boost::thread_group threadGroup;
  boost::asio::io_service::work work(ioService);

  //the we add the threads to the pool
  std::size_t numThreads = static_cast<std::size_t>(boost::thread::hardware_concurrency());
  for (std::size_t i = 0; i < numThreads; ++i)
  {
    threadGroup.create_thread(boost::bind(&boost::asio::io_service::run, &ioService));
  }

  std::string query = "SELECT * FROM " + inputDataAccess->getDataSetName() + " ORDER BY " + parentProperties;
  std::unique_ptr<te::da::DataSet> inputDataSet = inputDataAccess->query(query);

  //we create 10 tiles for each thread based on the dataSetSize
  std::size_t dataSetSize = inputDataSet->size();
  std::size_t jobSize = static_cast<std::size_t>(std::ceil(static_cast<double>(dataSetSize) / static_cast<double>(numThreads) / static_cast<double>(10)));
  if (jobSize == 0)
  {
    jobSize = dataSetSize;
  }

  if (m_impl->m_taskProgress != nullptr)
  {
    m_impl->m_taskProgress->setMessage(TE_TR("Finalizing - Merging data"));
    m_impl->m_taskProgress->setTotalSteps(static_cast<int>(dataSetSize));
    m_impl->m_taskProgress->setCurrentStep(0);
    m_impl->m_taskProgress->useMultiThread(true);
  }

  std::vector<std::size_t> vecParentPropertyIndexes;
  std::unique_ptr<te::mem::DataSet> outputDataSet(new te::mem::DataSet(outputDataSetType));

  std::vector<te::vp::FeatureSet> fsTile;
  fsTile.reserve(jobSize);

  te::vp::FeatureSet currentFeatureSet;
  std::string previousKey;
  inputDataSet->moveBeforeFirst();
  while (inputDataSet->moveNext() == true)
  {
    std::unique_ptr<te::da::DataSetItem> inputItem = inputDataSet->getItem();
    std::unique_ptr<te::vp::Feature> inputFeature(new te::vp::Feature(*inputItem.get()));

    const te::gm::Geometry* inputGeometry = inputFeature->getFirstGeometryPtr();
    if (inputGeometry == nullptr)
    {
      continue;
    }

    inputFeature->setSRID(srid);

    if (vecParentPropertyIndexes.empty() == true)
    {
      for (std::size_t i = 0; i < vecParentProperties.size(); ++i)
      {
        std::size_t pos = inputFeature->getPropertyPos(vecParentProperties.at(i));
        vecParentPropertyIndexes.push_back(pos);
      }
    }

    std::string currentKey;
    for (std::size_t i = 0; i < vecParentPropertyIndexes.size(); ++i)
    {
      if (currentKey.empty() == false)
      {
        currentKey += ",";
      }
      std::size_t pos = vecParentPropertyIndexes.at(i);
      if (inputFeature->isNull(pos) == true)
      {
        currentKey += "null";
      }
      else
      {
        currentKey += inputFeature->getValue(pos)->toString();
      }
    }

    if ((previousKey.empty() == true) || (previousKey == currentKey))
    {
      //we must just add to the current vector
      currentFeatureSet.push_back(inputFeature.release());
    }
    else
    {
      //we must apply the dissolve
      if (currentFeatureSet.empty() == false)
      {
        fsTile.push_back(currentFeatureSet);
        currentFeatureSet = te::vp::FeatureSet();
      }

      //we must just add to the current vector
      currentFeatureSet.push_back(inputFeature.release());
    }

    previousKey = currentKey;

    if (fsTile.size() == jobSize)
    {
      //adds the jobs to the manager
      ioService.post(boost::bind(DissolveMultiThread, outputDataAccess, fsTile, m_impl->m_taskProgress));
      fsTile = std::vector<te::vp::FeatureSet>();
      fsTile.reserve(jobSize);
    }
  }

  if (currentFeatureSet.empty() == false)
  {
    fsTile.push_back(currentFeatureSet);
  }

  if (fsTile.empty() == false)
  {
    //adds the jobs to the manager
    ioService.post(boost::bind(DissolveMultiThread, outputDataAccess, fsTile, m_impl->m_taskProgress));
    fsTile = std::vector<te::vp::FeatureSet>();
  }

  //wait for the jobs to be finished
  ioService.stop();
  threadGroup.join_all();

  return outputDataAccess;
}

std::vector<te::da::DataAccess*> te::vp::SubdivideByBorderTillingStrategy::merge(const std::vector<te::da::DataAccess*>& vecInputDataAccess)
{
  std::vector<te::da::DataAccess*> vecOutputDataAccess;
  //for each input DataAccessm we merge them individually
  for (std::size_t i = 0; i < vecInputDataAccess.size(); ++i)
  {
    te::da::DataAccess* inputDataAccess = vecInputDataAccess.at(i);
    te::da::DataAccess* outputDataAccess = merge(inputDataAccess);
    vecOutputDataAccess.push_back(outputDataAccess);
  }

  return vecOutputDataAccess;
}

te::da::DataAccess* te::vp::SubdivideByBorderTillingStrategy::createInputDataAccessForSubdivision(const te::da::DataSetType* baseOutputDataSetType, const std::string& dataSetName)
{
  std::string dataSetNameLower = te::common::Convert2LCase(dataSetName);

  std::unique_ptr<te::da::DataSetType> resultDataSetType(static_cast<te::da::DataSetType*>(baseOutputDataSetType->clone()));
  resultDataSetType->setName(dataSetNameLower);
  resultDataSetType->setDatasetName(dataSetNameLower);
  resultDataSetType->setTitle(dataSetNameLower);

  //we first add the subdivision columns
  appendSubdivideColumns(resultDataSetType.get(), dataSetNameLower);

  //the we create the data acceess
  te::da::DataAccess* outputDataAccess = CreateOutputDataAccess(resultDataSetType.get(), m_impl->createTempDataSource(), dataSetNameLower, m_impl->m_srid);

  //then we create the indexes
  createIndexes(outputDataAccess, dataSetNameLower);

  return outputDataAccess;
}

te::da::DataAccess* te::vp::SubdivideByBorderTillingStrategy::createOutputDataAccessForSubdivision(const te::da::DataSetType* baseOutputDataSetType, const std::vector<te::da::DataAccess*>& vecInputDataAccess, const std::string& dataSetName)
{
  std::unique_ptr<te::da::DataSetType> resultDataSetType(static_cast<te::da::DataSetType*>(baseOutputDataSetType->clone()));
  resultDataSetType->setName(dataSetName);
  resultDataSetType->setDatasetName(dataSetName);
  resultDataSetType->setTitle(dataSetName);

  //we first add the subdivision columns
  for (std::size_t i = 0; i < vecInputDataAccess.size(); ++i)
  {
    appendSubdivideColumns(resultDataSetType.get(), vecInputDataAccess.at(i)->getDataSetType()->getName());
  }

  //the we create the data acceess
  te::da::DataAccess* outputDataAccess = CreateOutputDataAccess(resultDataSetType.get(), m_impl->createTempDataSource(), dataSetName, m_impl->m_srid);

  //then we create the indexes
  for (std::size_t i = 0; i < vecInputDataAccess.size(); ++i)
  {
    createIndexes(outputDataAccess, vecInputDataAccess.at(i)->getDataSetType()->getName());
  }

  return outputDataAccess;
}

void te::vp::SubdivideByBorderTillingStrategy::createIndexes(te::da::DataAccess* dataAccess, const std::string& baseDataSetName)
{
  const te::da::DataSetType* dataSetType = dataAccess->getDataSetType();
  te::da::DataSourcePtr dataSource = dataAccess->getDataSource();
  std::string dataSetName = te::common::Convert2LCase(dataAccess->getDataSetName());

  std::string columnPropertyName = getFullPropertyName(baseDataSetName, this->getColumnPropertyName());
  std::string rowPropertyName = getFullPropertyName(baseDataSetName, this->getRowPropertyName());
  std::string parentPropertyName = getFullPropertyName(baseDataSetName, this->getParentPropertyName());

  {
    te::da::Index index("idx_" + dataSetName + "_" + columnPropertyName + "_" + rowPropertyName);
    index.add(dataSetType->getProperty(columnPropertyName));
    index.add(dataSetType->getProperty(rowPropertyName));
    std::map<std::string, std::string> mapOptions;
    dataSource->addIndex(dataSetName, &index, mapOptions);
  }

  {
    te::da::Index index("idx_" + dataSetName + "_" + parentPropertyName);
    index.add(dataSetType->getProperty(parentPropertyName));
    std::map<std::string, std::string> mapOptions;
    dataSource->addIndex(dataSetName, &index, mapOptions);
  }
}

std::string te::vp::SubdivideByBorderTillingStrategy::getColumnPropertyName() 
{
  std::string propertyName("subdivide_column");
  return propertyName;
}

std::string te::vp::SubdivideByBorderTillingStrategy::getRowPropertyName() 
{
  std::string propertyName("subdivide_row");
  return propertyName;
}

std::string te::vp::SubdivideByBorderTillingStrategy::getParentPropertyName()
{
  std::string propertyName("subdivide_parent");
  return propertyName;
}

std::string te::vp::SubdivideByBorderTillingStrategy::getFullPropertyName(const std::string& dataSetName, const std::string& propertyName)
{
  std::string fullPropertyName = dataSetName + "_" + propertyName;
  fullPropertyName = te::common::Convert2LCase(fullPropertyName);
  return fullPropertyName;
}

void te::vp::SubdivideByBorderTillingStrategy::appendSubdivideColumns(te::da::DataSetType* dataSetType, const std::string& baseDataSetName)
{
  te::dt::SimpleProperty* columnProperty = new te::dt::SimpleProperty(getFullPropertyName(baseDataSetName, getColumnPropertyName()), te::dt::INT32_TYPE);
  dataSetType->add(columnProperty);

  te::dt::SimpleProperty* rowProperty = new te::dt::SimpleProperty(getFullPropertyName(baseDataSetName, getRowPropertyName()), te::dt::INT32_TYPE);
  dataSetType->add(rowProperty);

  te::dt::SimpleProperty* parentIndexProperty = new te::dt::SimpleProperty(getFullPropertyName(baseDataSetName, getParentPropertyName()), te::dt::INT32_TYPE);
  dataSetType->add(parentIndexProperty);
}
