/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/AbstractDataHandler.cpp

   \brief This class represents handler to help reading and writing data from/to a dataSource by dividing it into smaller groups
 */

#include "SubdivideDataHandler.h"

#include "../common/progress/TaskProgress.h"
#include "../common/STLUtils.h"
#include "../core/translator/Translator.h"
#include "../datatype/AbstractData.h"
#include "../datatype/SimpleData.h"
#include "../dataaccess/dataset/DataSetAdapter.h"
#include "../dataaccess/dataset/DataSetTypeConverter.h"
#include "../dataaccess/utils/DataAccess.h"
#include "../dataaccess/utils/BasicQueries.h"
#include "../dataaccess/utils/DataAccess.h"
#include "../dataaccess/utils/Utils.h"
#include "../geometry/Envelope.h"
#include "../memory/DataSet.h" 
#include "../memory/PreparedItemCopy.h" 
#include "Feature.h"
#include "SubdivideByBorderTillingStrategy.h"
#include "Utils.h"

//Boost include files
#include <boost/thread/lock_guard.hpp>
#include <boost/thread/mutex.hpp>

namespace
{
  static te::vp::SubdivideDataHandlerFactory SubdivideDataHandlerFactoryInstance;
}


class SubdivideDataHandlerImpl
{
public:

  SubdivideDataHandlerImpl() {}
  ~SubdivideDataHandlerImpl() = default;

  std::vector<te::da::DataAccess*> m_vecOriginalInputData;
  std::vector<te::da::DataAccess*> m_vecOriginalOutputData;
  std::vector< std::vector<std::unique_ptr<te::dt::AbstractData>> > m_vecGroups;
  std::unique_ptr<te::vp::SubdivideByBorderTillingStrategy> m_strategy;
  boost::mutex m_mutex;
};


te::vp::SubdivideDataHandler::SubdivideDataHandler()
  : AbstractDataHandler()
  , m_impl(new SubdivideDataHandlerImpl())
{
}

//declaration just for the unique_ptr to be deletedon the cpp file
te::vp::SubdivideDataHandler::~SubdivideDataHandler() = default;

void te::vp::SubdivideDataHandler::initialize(const std::vector<te::da::DataAccess*>& vecInputData, const std::vector<te::da::DataAccess*>& vecOutputData, std::size_t maxCoordinates, te::common::TaskProgress* taskProgress)
{
  AbstractDataHandler::initialize(vecInputData, vecOutputData, maxCoordinates, taskProgress);

  m_impl->m_vecOriginalInputData = vecInputData;
  m_impl->m_vecOriginalOutputData = vecOutputData;

  int srid = m_impl->m_vecOriginalInputData.at(0)->getSRID();
  te::gm::Envelope fullEnvelope;
  for (std::size_t i = 0; i < m_impl->m_vecOriginalInputData.size(); ++i)
  {
    int currentSRID = m_impl->m_vecOriginalInputData.at(i)->getSRID();
    te::gm::Envelope currentEnvelope = m_impl->m_vecOriginalInputData.at(i)->getMBR();

    if (currentSRID != srid)
    {
      currentEnvelope.transform(currentSRID, srid);
    }
    
    fullEnvelope.Union(currentEnvelope);
  }

  std::size_t numColumns = 20;
  std::size_t numRows = 20;

  double resolutionX = fullEnvelope.getWidth() / static_cast<double>(numColumns);
  double resolutionY = fullEnvelope.getHeight() / static_cast<double>(numRows);

  //creates the strategy

  m_impl->m_strategy.reset(new te::vp::SubdivideByBorderTillingStrategy(fullEnvelope, srid, resolutionX, resolutionY, taskProgress));

  //subdivides
  m_vecInputData = m_impl->m_strategy->subdivide(m_impl->m_vecOriginalInputData, maxCoordinates);
  m_vecOutputData.clear();

  //prepares the subdivided output data access
  m_vecOutputData.push_back(m_impl->m_strategy->createOutputDataAccessForSubdivision(m_impl->m_vecOriginalOutputData.at(0)->getDataSetType(), m_vecInputData, "result"));

  //then we create the group list
  for (std::size_t row = 0; row < numRows; ++row)
  {
    for (std::size_t column = 0; column < numColumns; ++column)
    {
      std::unique_ptr<te::dt::AbstractData> rowData(new te::dt::Int32(static_cast<boost::int32_t>(row)));
      std::unique_ptr<te::dt::AbstractData> columnData(new te::dt::Int32(static_cast<boost::int32_t>(column)));

      std::vector<std::unique_ptr<te::dt::AbstractData>> vecKeys;
      vecKeys.push_back(std::move(rowData));
      vecKeys.push_back(std::move(columnData));

      m_impl->m_vecGroups.push_back(std::move(vecKeys));
    }
  }
}

void te::vp::SubdivideDataHandler::finalize(te::common::TaskProgress* taskProgress)
{
  //now we must merge the tiles
  std::vector<te::da::DataAccess*> vecMergedDataAccess = m_impl->m_strategy->merge(m_vecOutputData);

  te::mem::PreparedItemCopy copy;

  //and then copy the merge data access to the output dataccess
  for (std::size_t i = 0; i < vecMergedDataAccess.size(); ++i)
  {
    te::da::DataSourcePtr dataSource = vecMergedDataAccess.at(i)->getDataSource();
    std::unique_ptr<te::da::DataSet> dataSet = vecMergedDataAccess.at(i)->getDataSet();

    //check for empty results
    if (dataSet == nullptr)
    {
      continue;
    }

    const te::da::DataSetType* inputDataSetType = vecMergedDataAccess.at(i)->getDataSetType();
    te::da::DataSetType* outputDataSetTypeWithoutKey = static_cast<te::da::DataSetType*>(m_impl->m_vecOriginalOutputData.at(i)->getDataSetType()->clone());

    te::da::PrimaryKey* primaryKey = outputDataSetTypeWithoutKey->getPrimaryKey();
    if (primaryKey != nullptr)
    {
      if (primaryKey->getProperties().empty() == false)
      {
        const std::string& pkName = primaryKey->getProperties().at(0)->getName();
        te::dt::Property* keyProperty = outputDataSetTypeWithoutKey->getProperty(pkName);
        if (keyProperty != nullptr)
        {
          outputDataSetTypeWithoutKey->remove(keyProperty);
        }
      }
    }

    if (taskProgress != nullptr)
    {
      taskProgress->setMessage(TE_TR("Saving Results"));
      taskProgress->setTotalSteps(static_cast<int>(0));
      taskProgress->useMultiThread(true);
    }

    std::unique_ptr<te::mem::DataSet> outputDataSet(new te::mem::DataSet(outputDataSetTypeWithoutKey));

    dataSet->moveBeforeFirst();
    while (dataSet->moveNext())
    {
      te::vp::Feature* feature = new te::vp::Feature(outputDataSetTypeWithoutKey);
      copy.copyValues(dataSet->getItem().get(), feature);

      outputDataSet->add(feature);
    }

    if (outputDataSet->size() > 0)
    {
      m_impl->m_vecOriginalOutputData.at(i)->saveDataSet(outputDataSet.release(), true);
    }
  }

  te::common::FreeContents(vecMergedDataAccess);

  for (std::size_t i = 0; i < m_vecInputData.size(); ++i)
  {
    delete m_vecInputData.at(i);
  }

  for (std::size_t i = 0; i < m_vecOutputData.size(); ++i)
  {
    delete m_vecOutputData.at(i);
  }
}


std::size_t te::vp::SubdivideDataHandler::getGroupCount() const
{
  return m_impl->m_vecGroups.size();
}


te::vp::FeatureSet te::vp::SubdivideDataHandler::getFeatureSet(std::size_t groupIndex, std::size_t inputIndex, const te::gm::Envelope& filter)
{
  std::unique_ptr<te::da::DataSet> dataSet;

  const std::vector<std::unique_ptr<te::dt::AbstractData>>& vecKeys = m_impl->m_vecGroups.at(groupIndex);
  te::da::DataAccess* dataAccess = m_vecInputData.at(inputIndex);

  const te::da::DataSetType* dataSetType = dataAccess->getDataSetType();

  std::vector<std::string> vecColumns;
  for (std::size_t i = 0; i < dataSetType->size(); ++i)
  {
    vecColumns.push_back(dataSetType->getProperty(i)->getName());
  }

  std::string columnPropertyName = te::vp::SubdivideByBorderTillingStrategy::getFullPropertyName(dataSetType->getName(), te::vp::SubdivideByBorderTillingStrategy::getColumnPropertyName());
  std::string rowPropertyName = te::vp::SubdivideByBorderTillingStrategy::getFullPropertyName(dataSetType->getName(), te::vp::SubdivideByBorderTillingStrategy::getRowPropertyName());

  std::vector<std::string> vecGroupColumns;
  vecGroupColumns.push_back(columnPropertyName);
  vecGroupColumns.push_back(rowPropertyName);

  std::unique_ptr<te::da::Select> selectGroup = te::da::BasicQueries::GetQuery(false, vecColumns, m_vecInputData.at(inputIndex)->getDataSetName(), vecGroupColumns, vecKeys);

  //creates a scope so the locker can be used just in the query of data
  {
    boost::lock_guard<boost::mutex> lockGuard(m_impl->m_mutex);

    dataSet = dataAccess->query(*selectGroup);
  }

  //declares the resulting featureset
  te::vp::FeatureSet featureSet;
  if (dataSet == nullptr)
  {
    return featureSet;
  }

  dataSet->moveBeforeFirst();
  while (dataSet->moveNext())
  {
    te::vp::Feature* feature = new te::vp::Feature(*dataSet->getItem().get());
    featureSet.push_back(feature);
  }

  return featureSet;
}

void te::vp::SubdivideDataHandler::addFeatureSet(std::size_t groupIndex, std::size_t outputIndex, const te::vp::FeatureSet& featureSet)
{
  if (featureSet.empty() == true)
  {
    return;
  }

  te::da::DataAccess* dataAccess = m_vecOutputData.at(outputIndex);
  std::unique_ptr<te::da::DataSet> dataSet = te::vp::CreateDataSetFromFeatureSet(dataAccess->getDataSetName(), dataAccess->getDataSetType(), featureSet);

  //the lock is only used in the saving of the geometries
  boost::lock_guard<boost::mutex> lockGuard(m_impl->m_mutex);
  dataAccess->saveDataSet(dataSet.release(), false);
}


te::vp::SubdivideDataHandlerFactory::SubdivideDataHandlerFactory()
  : te::vp::AbstractDataHandlerFactory("SubdivideDataHandler")
{

}

te::vp::SubdivideDataHandlerFactory::~SubdivideDataHandlerFactory() = default;

te::vp::AbstractDataHandler* te::vp::SubdivideDataHandlerFactory::build()
{
  return new te::vp::SubdivideDataHandler();
}
