/*  Copyright (C) 2008 National Institute For Space Research (INPE) - Brazil.

    This file is part of the TerraLib - a Framework for building GIS enabled applications.

    TerraLib is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License,
    or (at your option) any later version.

    TerraLib is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with TerraLib. See COPYING. If not, write to
    TerraLib Team at <terralib-team@terralib.org>.
 */

 /*!
   \file terralin/vp/SubdivideOperation.cpp

   \brief Represents a make valid operation which analyses and fixes features of a set
 */
#include "SubdivideOperation.h"

#include "../datatype/SimpleProperty.h"
#include "../dataaccess/dataset/DataSetType.h"
#include "../geometry/Geometry.h"
#include "../geometry/GeometryCollection.h"
#include "../geometry/Utils.h"
#include "../memory/PreparedItemCopy.h"
#include "CommonFunctions.h"
#include "Feature.h"
#include "PolygonSubdivider.h"

namespace te
{
  namespace vp
  {
    namespace subdivideOperation
    {
      const std::string OPERATION_NAME = "Subdivide";
    }
  }
}

namespace
{
  static te::vp::SubdivideOperationFactory
    SubdivideOperationFactoryInstance;
}

te::vp::SubdivideOperationParameters::SubdivideOperationParameters(std::size_t maxCoordinates)
  : te::vp::AbstractParameters()
  , m_maxCoordinates(maxCoordinates)
{

}
te::vp::SubdivideOperationParameters::~SubdivideOperationParameters() = default;

void te::vp::SubdivideOperationParameters::setMaxCoordinates(std::size_t maxCoordinates)
{
  m_maxCoordinates = maxCoordinates;
}

std::size_t te::vp::SubdivideOperationParameters::getMaxCoordinates() const
{
  return m_maxCoordinates;
}

te::vp::SubdivideOperation::SubdivideOperation()
    : AbstractOperation(te::vp::subdivideOperation::OPERATION_NAME)
    , m_parentIdColumnIndex(std::string::npos)
    , m_parentId(0)
{

}

te::vp::SubdivideOperation::~SubdivideOperation() = default;

std::unique_ptr<te::vp::AbstractOperationCapabilities> te::vp::SubdivideOperation::getCapabilities() const
{
  std::unique_ptr<te::vp::AbstractOperationCapabilities> capabilities(new te::vp::AbstractOperationCapabilities(getName(), 1, 1, "UnorderedGroupingDataHandler", false));
  return capabilities;
}

std::vector<te::vp::FeatureSet> te::vp::SubdivideOperation::executeImpl(const std::vector<te::vp::FeatureSet>& vecInput)
{
  std::vector<te::vp::FeatureSet> vecOutput;
  if (vecInput.size() != 1)
  {
    vecOutput.resize(1);
    return vecOutput;
  }

  std::size_t maxCoordinates = 0;
  te::vp::SubdivideOperationParameters* subdivideParameters = dynamic_cast<te::vp::SubdivideOperationParameters*>(this->getParameters());
  if (subdivideParameters != nullptr)
  {
    maxCoordinates = subdivideParameters->getMaxCoordinates();
  }

  const te::vp::FeatureSet& inputFeatureSet = vecInput.at(0);

  te::mem::PreparedItemCopy copy;

  te::vp::FeatureSet outputFeatureSet;

  int srid = inputFeatureSet.at(0)->getFirstGeometryPtr()->getSRID();
  double precision = te::gm::GetAppropriatePrecision(srid);

  //we must convert the geometries to the Identity operation output type
  for (std::size_t i = 0; i < inputFeatureSet.size(); ++i)
  {
    const te::vp::Feature* inputFeature = inputFeatureSet.at(i);
    const te::gm::Geometry* geometry = inputFeature->getFirstGeometryPtr();
    if (geometry == nullptr)
    {
      continue;
    }

    if (this->isAborted() == true)
    {
      vecOutput.push_back(outputFeatureSet);
      return vecOutput;
    }

    const te::gm::Envelope* envelope = geometry->getMBR();
    double x1 = envelope->m_llx - precision;
    double y1 = envelope->m_lly - precision;
    double x2 = envelope->m_urx + precision;
    double y2 = envelope->m_ury + precision;

    te::gm::Envelope subEnvelope(x1, y1, x2, y2);

    double area1 = te::gm::GetArea(geometry);

    std::size_t parentId = safeCreateParentId();

    te::gm::GeometryVector vecSingleOutput = te::vp::PolygonSubdivider::subdivide(geometry, subEnvelope, maxCoordinates);
    for (std::size_t j = 0; j < vecSingleOutput.size(); ++j)
    {
      te::gm::GeometryPtr outputGeometry(vecSingleOutput.at(j));

      te::gm::GeometryVector vecAdjustedOutput = te::vp::AdjustToGeomType(vecSingleOutput.at(j), getOutputGeometryType());
      for (std::size_t k = 0; k < vecAdjustedOutput.size(); ++k)
      {
        te::gm::GeometryPtr adjustedOutputGeometry(vecAdjustedOutput.at(k));

        te::vp::Feature* convertedFeature = createOutputFeature(0);
        copy.copyValues(inputFeature, convertedFeature, false);
        convertedFeature->setFirstGeometry(adjustedOutputGeometry.release());

        if (m_parentIdColumnIndex != std::string::npos)
        {
          convertedFeature->setInt32(m_parentIdColumnIndex, static_cast<boost::int32_t>(parentId));
        }

        outputFeatureSet.push_back(convertedFeature);
      }
    }
  }

  vecOutput.push_back(outputFeatureSet);
  
  return vecOutput;
}

te::da::DataSetType* te::vp::SubdivideOperation::createBasicOutputDataSetType(const std::string& dataSetName, int srid, te::gm::GeomType geometryType, const std::string& geometryColumnName) const
{
  te::da::DataSetType* basicDataSetType = AbstractOperation::createBasicOutputDataSetType(dataSetName, srid, geometryType, geometryColumnName);
  {
    te::dt::SimpleProperty* absProperty = new te::dt::SimpleProperty("subdivide_parent_id", te::dt::INT32_TYPE, true);
    basicDataSetType->add(absProperty);

    m_parentIdColumnIndex = basicDataSetType->getPropertyPosition("subdivide_parent_id");
  }

  return basicDataSetType;
}

std::size_t te::vp::SubdivideOperation::safeCreateParentId()
{
  const std::lock_guard<std::mutex> lock(m_mutex);
  ++m_parentId;

  return m_parentId;
}

te::vp::SubdivideOperationFactory::SubdivideOperationFactory()
  : OperationFactory(te::vp::subdivideOperation::OPERATION_NAME)
{
}

te::vp::SubdivideOperationFactory::~SubdivideOperationFactory() = default;

te::vp::AbstractOperation* te::vp::SubdivideOperationFactory::build()
{
  return new SubdivideOperation();
}
